/* 
 * Alto debugger 
 * */



#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <readline/readline.h>
#include "global.h"
#include "functions.h"
#include <excpt.h>
#include <filehdr.h>
#include <sym.h>
#include <symconst.h>
#include "flags.h"
#include "regs.h"

int cbreaks=0;
char ** breaks=NULL;
char * inbuffer=NULL;
long global_ask=0;
#define ADD_BREAK(naam)	add_break(naam)

/* FUNCTIONS */

void bbl2fun();
void cont();
void help();
void printfun();
void printloops();
void printvaluespec();
void setbreak();
void showfun();
void off();
void on();
void quit();
void clearranges();

/* */
typedef struct range 
{
   char * f;
   long a;
   long b;
} r_type;

r_type * ranges=NULL;
long cranges=0;

int askit(char * s,INDEX i)
{
	char string[200];
	char string2[200];
	char * out=s;
	long tel;
	long low=-1;
	long high=-1;

	for (tel=0; tel<cranges; cranges++)
	{
	   if (ranges[tel].f==s)
	   {
	      if ((ranges[tel].a<=i)&& (ranges[tel].b>=i))
	      {
		 printf("Doing %d\n",i);
		 return 1;
	      }
	      return 3;
	   }
	}
	
	while (1)
	{
	   free(inbuffer);
	   if (i!=-1)
	   {
	     sprintf(string2,s,i);
	     out=string2; 
	   }
	   if (VALID(i))
	   {
	      sprintf(string,"Do %s [Y/N/A(lways)/D(on't even think about it)/R(ange)]: ",out);
	   }
	   else
	      sprintf(string,"Do %s [Y/N/A(lways)/D(on't even think about it)]: ",out);
	   inbuffer=readline(string);
	   if (inbuffer==NULL) continue;
	   if (inbuffer[0]=='Y')
	   {
	      printf("Yes\n");	
	      return 1;
	   }
	   if (inbuffer[0]=='N') 
	   {
	      printf("No\n");	
	      return 3;
	   }
	   if (inbuffer[0]=='A') 
	   {
	      printf("Always\n");
	      return 2;
	   }
	   if (inbuffer[0]=='D') 
	   {
	      printf("Never\n");
	      return 0;
	   }
	   if (inbuffer[0]=='R') 
	   {
	      sscanf(inbuffer+1,"%ld %ld\n",&low,&high);	
	      if ((low<0)||(high<0))
	      {
		 printf("Invalid range %ld %ld\n",low,high);
	      }
	      else
	      {
		 printf("Doing range\n");
		 ranges=realloc(ranges,sizeof(r_type)*(cranges+1));
		 ranges[cranges].f=s;
		 ranges[cranges].a=low;
		 ranges[cranges].b=high;
		 cranges++;
		 if ((low<=i)&& (high>=i))
		    return 1;
		 return 3;
	      }
	      printf("OK\n");
	   }
	}
}

void clearranges()
{
   free(ranges);
   cranges=0;
}

void debug(char * filename, int regel);




void add_break(char * naam)
{
	breaks=realloc(breaks,sizeof(char *)*(cbreaks+1));
	breaks[cbreaks]=malloc(sizeof(char)*strlen(naam)+1);
	strcpy(breaks[cbreaks],naam);
	cbreaks++;
}

static int debugstop=1;
static int interactive=1;
static int offtil=-1;
static int monitor=-1;
static int fileteller=0;
static char * checknaam=NULL;
int debug_arg=-1;
int outfd;
int infd=-1;
char * breakpoint=NULL;

int functionnumber(char * tmp);
void dumpfun(FILE * output, INDEX ifun);
void dumpfunnon(FILE * output, INDEX ifun);

typedef struct {
	  char *name;                   /* User printable name of the function. */
	  char *sh_name;                   /* User printable name of the function. */
	  void (*fun)();			/* Function to call */
	  char *doc;                    /* Documentation for this function.  */
} COMMAND;

COMMAND commands[] = {
	{ "bbl2fun","b2f", bbl2fun,"Given a bbl give the function" },
	{ "break","b", setbreak,"break at an option check" },
	{ "clearranges","cr", clearranges,"clears all askranges" },
	{ "cont", "c", cont, "Continue" },
	{ "help", "h", help, "Print help" },
	{ "printfun", "pf", printfun,"Print a function" },
	{ "showfun", "sf", showfun,"Show a function" },
	{ "printloops", "pl", printloops,"Print loops of a function" },
	{ "printvaluespec", "pv", printvaluespec,"Print Incoming ValueSpec of a bbl" },
	{ "off","off", off,"Turns debugging off" },
	{ "on", "on", on,"Turns debugging on" },
	{ "quit", "q", quit,"Exit Alto" },
	{ (char *)NULL, (char *) NULL, NULL,(char *)NULL }
};



void init_debug()
{
#include "breakpoints.h"
}

char * dupstr (char * s)
{
	char *r;

	r = malloc (sizeof(char)*(strlen (s) + 1));
	strcpy (r, s);
	return (r);
}


char * command_generator (char *text, int state)
{
	static int list_index, len;
	char *name;

	if (!state)
	{
		list_index = 0;
		len = strlen (text);
	}

	while ((name=commands[list_index++].name))
		if (strncmp (name, text, len) == 0)
			return dupstr(name);

	return ((char *)NULL);
}

char * fun_generator (char * text, int state)
{
	static int list_index, len;
	char *name;
	INDEX ifun;

	if (!state)
	{
		list_index = 0;
		len = strlen (text);
	}

	FOREACH_LIVE_FUN(ifun)
	{
		if (list_index>=ifun) continue;
		name=AFUN_NAME(ifun);
		list_index=ifun;
		if (strncmp (name, text, len) == 0)
		{
		   return dupstr(name);
		}
	}

	return ((char *)NULL);
}

char *
break_generator (text, state)
	char *text;
	int state;
{
	static int list_index, len;
	char *name;
	char *ret;

	/* If this is a new word to complete, initialize now.  This includes
	 *      saving the length of TEXT for efficiency, and initializing the index
	 *           variable to 0. */
	if (!state)
	{
		list_index = 0;
		len = strlen (text);
	}

	/* Return the next name which partially matches from the command list. */
	for (; list_index<cbreaks; list_index++)	
	{
		name=breaks[list_index];
		if (strncmp (name, text, len) == 0)
		{
			ret=dupstr(name);
			list_index++;
			return ret;
		}
	}

	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}


char **
my_completion (text, start, end)
	     char *text;
	          int start, end;
{
	char **matches;

	matches = (char **)NULL;

	if (start == 0)
		matches = completion_matches (text, command_generator);
	else
	{
		if (strncmp(rl_line_buffer,"printfun ",9)==0) matches = completion_matches (text, fun_generator);
		if (strncmp(rl_line_buffer,"pf ",3)==0) matches = completion_matches (text, fun_generator);
		if (strncmp(rl_line_buffer,"printloops ",11)==0) matches = completion_matches (text, fun_generator);
		if (strncmp(rl_line_buffer,"pl ",3)==0) matches = completion_matches (text, fun_generator);
		if (strncmp(rl_line_buffer,"break ",6)==0) matches = completion_matches (text, break_generator);
		if (strncmp(rl_line_buffer,"b ",2)==0) matches = completion_matches (text, break_generator);

	}

	return (matches);
}




void check_debug(char * str)
{
	if (str==NULL) return;
	if (breakpoint==NULL) return;
	if (strcmp(str,breakpoint)!=0) return;
	debug("Breakpoint",1);
}

void check_changes()
{
	static long savechecksum=-1;
	long checksum=1;
	INDEX ibbl,iins;
	int teller=0;
	if (monitor==-1) return;
	FOREACH_FUN_BBL(monitor,ibbl)
	{
		teller++;
		checksum+=ibbl*teller*AbblNInss(ibbl);
		FOREACH_BBL_INS(ibbl,iins)
		{
			checksum+=AINS_CODE(iins);
		}
	}
	if (checksum!=savechecksum)
	{
		if (checknaam!=NULL)
		{
			char  tmpnaam[256];
			FILE * file;
			sprintf(tmpnaam,"%s-%d",checknaam,fileteller);
			file=fopen(tmpnaam,"w");
			fprintf(file, "\n%%StartSloop\n");
			PrintFunLoops(file,monitor,1);
			fprintf(file, "\n%%EndSloop\n");
			fclose(file);
			fileteller++;
		}
		else
		dumpfun(stdout,monitor);
		savechecksum=checksum;
	}
}

extern char * exec_name;
void print_call_chain(int i)
{
	FILE * fp=fopen(exec_name,"r");
	FILHDR fh;
	long bound=0;
	int sym=-1;
	int tel;
	HDRR symtab;
	char name[100];
	EXTR * table;
	CONTEXT contextRecord;
	PRUNTIME_FUNCTION prf; 

	if (fp==0)
	  {
	    printf("Execute squeeze using its full path to see the call stack\n");
	    return;
	  }
	printf("---- Call Stack ----\n");
	exc_capture_context(&contextRecord);
	prf=exc_lookup_function_entry(contextRecord.sc_pc);
	exc_virtual_unwind(prf, &contextRecord);
	fread(&fh,sizeof(FILHDR),1,fp);
	fseek(fp,fh.f_symptr,SEEK_SET);
	fread(&symtab,sizeof(HDRR),1,fp);
	if (symtab.magic!=magicSym) return;
	table=malloc(sizeof(EXTR)*symtab.iextMax);
	fseek(fp,symtab.cbExtOffset,SEEK_SET);
	fread(table,sizeof(EXTR),symtab.iextMax,fp);
	do
	{
	for (tel=0; tel<symtab.iextMax; tel++)
	{
		if (table[tel].asym.st==stProc)
		{
			if ((table[tel].asym.value<contextRecord.sc_pc)&&(table[tel].asym.value>bound))
			{
				bound=table[tel].asym.value;
				sym=tel;
			}
		}
	}
	if (sym==-1) return;
	fseek(fp,symtab.cbSsExtOffset+table[sym].asym.iss,SEEK_SET);
	fread(name,sizeof(char),100,fp);
	printf("Name: %s %lx\n",name,contextRecord.sc_pc);
	
	prf=exc_lookup_function_entry(contextRecord.sc_pc);
	exc_virtual_unwind(prf, &contextRecord);
	bound=0;
	sym=-1;
	
	} while(strcmp(name,"main")!=0);

}


extern void isignal(int );
void debug(char * filename, int regel)
{
char buffer[256];
int number;
char * tmp;
outfd=fileno(stdout);
infd=fileno(stdin);
if (!debugstop) return;
if (offtil!=-1)
{
	if (debug_arg!=offtil) return;
}

offtil=-1;

sprintf(buffer,"Debug called in %s at line %d\n",filename,regel);
print_call_chain(0);
write(fileno(stdout),buffer,strlen(buffer));
while(1)
{
	int tel=0;
	free(inbuffer);
	rl_attempted_completion_function = (CPPFunction *)my_completion;
	inbuffer=readline("Alto debug: ");
	add_history(inbuffer);
	
	/* DECODE */

	tmp=strtok(inbuffer," \n");
	if (tmp==NULL) continue;

	while(commands[tel].name)
	{
		if ((strcmp(tmp,commands[tel].name)==0) || (strcmp(tmp,commands[tel].sh_name)==0))
		{
			if (commands[tel].fun!=NULL)
				commands[tel].fun();
			else
				printf("No such function .... \n");
			break;
		}
		tel++;
	}
	if ((strcmp(tmp,"cont")==0)||(strcmp(tmp,"off")==0)||(strcmp(tmp,"c")==0))
		return;
	if (commands[tel].name==NULL)		
	printf("Unknown command %s\n",tmp);

	continue;
	if (strcmp(tmp,"printfuns")==0) 
	{
		FOREACH_FUN(number)
		{
		write(outfd,AFUN_NAME(number),strlen(AFUN_NAME(number)));
		if (AFUN_IS_KILLED(number))
		write(outfd,"{Killed}",8);	
		write(outfd,"\n",1);	
		}
		write(outfd,"EOF\n",5);	
		continue;
	}

	if (strcmp(tmp,"printnums")==0) 
	{
		tmp=strtok(NULL," \n");
		FOREACH_FUN(number)
		{
			if (AFUN_IS_KILLED(number)) continue;
			if (strcmp(tmp,AFUN_NAME(number))==0)
			{
				sprintf(buffer,"%d\n",number);
				write(outfd,buffer,strlen(buffer));
			}
		}
		write(outfd,"EOF\n",5);	
		continue;
	}

	if (strcmp(tmp,"printlivefuns")==0) 
	{
		FOREACH_FUN(number)
		{
		if (AFUN_IS_KILLED(number)) continue;
		write(outfd,AFUN_NAME(number),strlen(AFUN_NAME(number)));
		sprintf(buffer,"(%d)",AfunNInss(number));
		write(outfd,buffer,strlen(buffer));
		write(outfd," ",1);	
		}
		write(outfd,"EOF\n",5);	
		continue;
	}

	if (strcmp(tmp,"offtilfun")==0) 
	{
		tmp=strtok(NULL," \n");
		number=functionnumber(tmp);
		if (!VALID(number)) continue;
		if (interactive)
		{
			sprintf(buffer,"Debug off until fun %s\n",tmp);
			write(outfd,buffer,strlen(buffer));
		}
		offtil=number;
		write(outfd,"EOF\n",5);	
		return;
	}
	if (strcmp(tmp,"change")==0) 
	{
		tmp=strtok(NULL," \n");
		number=functionnumber(tmp);
		if (!VALID(number)) continue;
		if (interactive)
		{
			sprintf(buffer,"Logging changes for fun %s\n",tmp);
			write(outfd,buffer,strlen(buffer));
		}

		tmp=strtok(NULL," \n");
		if (tmp!=NULL)
		{
			free(checknaam);
			fileteller=0;
			checknaam=(char *) malloc(sizeof(char)*(strlen(tmp)+1));
			strcpy(checknaam,tmp);
			checknaam[strlen(tmp)]=0;
		}
		else
		{
			free(checknaam);
			checknaam=NULL;
		}
		monitor=number;
		write(outfd,"EOF\n",5);	
		continue;
	}

}
}

/* HULPFUNKTIES */

void dumpfun(FILE * output, INDEX ifun)
{
	INDEX ibbl;
	INDEX iedge;
	fprintf(output,"digraph graaf {\n");
	FOREACH_FUN_BBL(ifun,ibbl)
	{
		if ABBL_IS_KILLED(ibbl) continue;
		FOREACH_BBL_SUCC(ibbl,iedge)
		{
			if (AEDG_IS_KILLED(iedge)) continue;
			if (AEDG_TYPE(iedge)==BT_RET) continue;
			if (!AEDG_IS_INTER(iedge))
			fprintf(output,"B%d -> B%d\n",ibbl,AEDG_TAIL(iedge));
			else
			fprintf(output,"B%d -> \"%s\"\n",ibbl,AFUN_NAME(ABBL_FUN(AEDG_TAIL(iedge))));
		}
	}
	fprintf(output,"}\n");
}

void dumpfunnon(FILE * output, INDEX ifun)
{
	INDEX ibbl;
	INDEX iedge;
	fprintf(output,"digraph graaf {\n");
	FOREACH_FUN_BBL(ifun,ibbl)
	{
		if ABBL_IS_KILLED(ibbl) continue;
		if (ABBL_COUNT(ibbl)==0) continue;
		FOREACH_BBL_SUCC(ibbl,iedge)
		{
			if (AEDG_IS_KILLED(iedge)) continue;
			if (AEDG_TYPE(iedge)==BT_RET) continue;
			if (!AEDG_IS_INTER(iedge))
			fprintf(output,"B%d -> B%d\n",ibbl,AEDG_TAIL(iedge));
			else
			fprintf(output,"B%d -> \"%s\"\n",ibbl,AFUN_NAME(ABBL_FUN(AEDG_TAIL(iedge))));
		}
	}
	fprintf(output,"}\n");
}

int functionnumber(char * tmp)
{
	int number;
	INDEX ifun;
	char buffer[256];
		number=0;
		if (atoi(tmp)!=0) number=atoi(tmp);

		if (number==0)
		FOREACH_FUN(ifun)
		{
			if (strcmp(tmp,AFUN_NAME(ifun))==0)
			{
				number=ifun;
				if (!(AFUN_IS_KILLED(ifun))) break;
			}
		}
	
		if (number==0) 
		{
			sprintf(buffer,"Function %s not found\n",tmp);
			write(outfd,buffer,strlen(buffer));
			return -1;
		}
		if (number<0 || number>CONFIG_UFUN)
		{
			sprintf(buffer,"Illegal function\n");
			write(outfd,buffer,strlen(buffer));
			return -1;
		}
		if (AFUN_IS_KILLED(number))
		{
			sprintf(buffer,"Function %s is killed\n",tmp);
			write(outfd,buffer,strlen(buffer));
			return -1;
		}
		return number;
}


/* IMPLEMENTATIE VAN ECHTE FUNCTIES */


void bbl2fun()
{
	char * tmp=NULL;
	int ibbl=0;
	tmp=strtok(NULL," \n");
	if (tmp!=NULL)
	{
		ibbl=atoi(tmp);
		if (!VALID(ibbl)) 
			printf("Bbl number not valid\n");
		else
		{
			if (ibbl>CONFIG_UBBL)
			{

				printf("No such bbl\n");
			}
			else
				if (ABBL_IS_KILLED(ibbl)) 
					printf("Bbl is killed\n");
				else
					printf("Function %s (%d)\n",AFUN_NAME(ABBL_FUN(ibbl)),ABBL_FUN(ibbl));
		}

	}
}

void setbreak()
{
	char * tmp;
	tmp=strtok(NULL," \n");
	if (tmp!=NULL)
	{
		free(breakpoint);
		breakpoint=(char *) malloc(sizeof(char)*(strlen(tmp)+1));
		strcpy(breakpoint,tmp);
		breakpoint[strlen(tmp)]=0;
	}

	printf("Breakpoint set\n");
}

void cont()
{
			printf("Continue\n");
}

void help()
{
		int tel=0;
		while (commands[tel].name)
		  {
		    printf("%14s: %s (shorthand: %s)\n",commands[tel].name,commands[tel].doc,commands[tel].sh_name);
		    tel++;
		  }
}

void printfun()
{
	char * tmp;
	INDEX number;
	tmp=strtok(NULL," \n");
	if (tmp==NULL)
	{
	   printf("Expected an argument....\n");
	   return;
	}
	number=functionnumber(tmp);
	if (!VALID(number)) return;
	printf("Printing function %s\n",tmp);
	tmp=strtok(NULL," \n");
	if (tmp!=NULL)
	{
		FILE * file=fopen(tmp,"w");
		PrintFunSuper(file,number);
		fclose(file);
	}
	else
		PrintFunSuper(stdout,number);
}

void showfun()
{
	char * tmp;
	INDEX number;
	tmp=strtok(NULL," \n");
	number=functionnumber(tmp);
	if (!VALID(number)) return;
	printf("Show function %s\n",tmp);
	//	doX(number);
}

void printvaluespec()
{
	char * tmp;
	INDEX number;
		
	tmp=strtok(NULL," \n");
	number=atoi(tmp);
	if (!VALID(number)) return; 
	printf("Printing Valuespec of Bbl %d\n",number);
	if (GetBblValueSpec(number)==NULL)
	{
	   printf("No valuespec for bbl %d\n",number);
	   return;
	}
	PrintValueSpec(GetBblValueSpec(number),REGS_ALL);
}


void printloops()
{
	char * tmp;
	INDEX number;
		
	tmp=strtok(NULL," \n");
	number=functionnumber(tmp);
	if (!VALID(number)) return; 
	printf("Printing Sloop of function %s\n",tmp);
	tmp=strtok(NULL," \n");
	if (tmp!=NULL)
	{
		FILE * file=fopen(tmp,"w");
		fprintf(file, "\n%%StartSloop\n");
		PrintFunLoops(file,number,1);
		fprintf(file, "\n%%EndSloop\n");
		fclose(file);
	}
	else
		PrintFunLoops(stdout,number,1);
}

void off()
{
  		signal(2,exit);
		debugstop=0;
}

void on()
{
  debugstop=1;
}
void quit()
{
		printf("Exit...\n");
		exit(0);
}


