#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"
#include "fltconst.h"
#include "blocks.h"

#define OFF 0
#define ON  1
#define CS 0  // call-site specific
#define CI 1  // call-site unspecific

extern int inoptdata;
int super_change;

/******************************************
 choose type of PE
******************************************/
#define PROPAGATION_METHOD                        CS  //call-site specific (CS) or unspecific (CI)

#define BBL_CONSTANT_PROPAGATION                  ON   //propagate over bbl-boundaries
#define PREDICATE_CONSTANT_PROPAGATION            ON //do backwards search for constants at branches
#define CONDITIONAL_CONSTANT_PROPAGATION          ON //do conditional CP instead of simple CP
#define INTERPROCEDURAL_CONSTANT_PROPAGATION      ON  //propagate over procedure boundaries
#define ELIMINATION                               ON  //do global OptConstGen's, eliminate idempotent instructions
#define LATE_CALL_JOIN                            ON  //do a separate propagation of incoming constants at call-sites, only meet the constants
                                                       //at the instructions instead of at the entry of the procedure and find new occasions
                                                       //idempotent (=useless) instructions and evaluable conditional branches
#define PROP_DEAD_VALUES                          ON  // propagate dead values as well
#define DETECT_DEAD_CALL_EDGES                    ON  // detects dead edges from AFUN_PSEUDO_HELL_LITE and does not propagate constants for them

/*******************************************
 macro definitions
*******************************************/

#define CONST(spec,reg)       (!((spec->bot | spec->top) & REG2BIT(reg)))


#define AEDG_LIVE(ibbl)      (AEDG_FLAGS(ibbl) &  FE_LIVE)
#define AEDG_UNLIVE(ibbl)     AEDG_FLAGS(ibbl) &= ~FE_LIVE
#define AEDG_SET_LIVE(ibbl)   AEDG_FLAGS(ibbl) |= FE_LIVE

#define ABBL_LIVE(ibbl)        ABBL_IS_VISITED(ibbl)
#define ABBL_UNLIVE(ibbl)      ABBL_UNSET_VISITED(ibbl)
#define ABBL_SET_LIVE(ibbl)    ABBL_SET_VISITED(ibbl)

INT64 TheLatticeTop = 0x3141592653589793;
INT64 TheLatticeBot = ~0x3141592653589793;

static void AbblSetValueSpec(INDEX ibbl);
static void AfunSetCI(INDEX ifun);
static void AfunConstProp(INDEX ifun);
#if PROPAGATION_METHOD
static BOOL AfunChangeCI(INDEX ifun);
#else 
static BOOL AfunChangeCS(INDEX ifun,INDEX call);
#endif
static void AbblMarkValueSpec(INDEX ibbl);
static void AfunSetCS(INDEX ifun, INDEX call);
static void AbblProvSetValueSpec(INDEX ibbl);
extern BOOL OptConstGenOneNew(FILE *fp,INDEX ibbl,INDEX iins,VALUESPEC * before);
extern BOOL OptConstGenTwoNew(FILE *fp,INDEX ibbl);

static void IntraEdgEvalValueSpec(INDEX iedg, VALUESPEC* output);
static BOOL IntraEdgProp2Succ(INDEX ibbl,VALUESPEC* bbl_output_spec);
static void AfunConstPropIntraEdg(INDEX ifun,INDEX call);
static BOOL AfunIntraEdgChangeCI(INDEX ifun);
static void AfunIntraEdgSetCI(INDEX ifun);
static BOOL AfunIntraEdgChangeCS(INDEX ifun, INDEX call);
static void AbblJoinIntraEdgValueSpecs(INDEX ibbl);

/*======================================================
  Mark a block for repropagation
  Used when a load was performed from data that now seems
  not to be constant anymore
 =======================================================*/

extern void AbblReprop(INDEX ibbl)
{
  INDEX iedg;
  AFUN_MARK(ABBL_FUN(ibbl));
#if !(PROPAGATION_METHOD)
  FOREACH_BBL_PRED(AFUN_BBL_FIRST(ABBL_FUN(ibbl)),iedg)
    {
      if (!AEDG_IS_INTER(iedg)) continue;
      AEDG_MARK(iedg);
    }
#endif
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "StrRedICM"

static NUMBER StrRedICM(INDEX iins)
{
  if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
      BOOL taken;

      switch(AINS_CODE(iins))
        {
        case I_cmoveq:   taken = AINS_A_EVAL(iins) ? 0 : 1; break;
	  
        case I_cmovne:   taken = AINS_A_EVAL(iins) ? 1 : 0; break;
	  
        case I_cmovlt:   taken = AINS_A_EVAL(iins) < 0 ? 1 : 0; break;
	  
        case I_cmovge:   taken = AINS_A_EVAL(iins) < 0 ? 0 : 1; break;
	  
        case I_cmovgt:   taken = AINS_A_EVAL(iins) > 0 ? 1 : 0; break;
	  
        case I_cmovle:   taken = AINS_A_EVAL(iins) > 0 ? 0 : 1; break;
	  
        case I_cmovlbs:  taken = AINS_A_EVAL(iins) & 1 ? 1 : 0; break;
	  
        case I_cmovlbc:  taken = AINS_A_EVAL(iins) & 1 ? 0 : 1; break;
	  
        default:
	  taken = 1;
	  ERROR(("Unexpected ICM instruction\n"));
	  PrintIns(stdout,iins);
	  break;
        }

      if( taken )
	{
	  if( AINS_REG_B(iins) == REG_LIT )
	    AinsMakeLda(iins,REG_ZERO,AINS_REG_C(iins),AINS_DATA(iins));
	  else
	    AinsMakeLda(iins,AINS_REG_B(iins),AINS_REG_C(iins),0);
	  return 1;
	}
      else
	{
	  AinsKill(iins);
	  return 1;
	}
	
    }
    
  if( AINS_REG_B(iins) == AINS_REG_C(iins) )
    {
      AinsKill(iins);
      return 1;
    }
    
  return 0;
}
#undef __PROC__
#define __PROC__ "StrRedFCM"

static NUMBER StrRedFCM(INDEX iins)
{
  if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
      BOOL taken;
      
      switch(AINS_CODE(iins))
        {
        case I_fcmoveq:   taken = AINS_A_EVAL(iins) == 0.0 ? 1 : 0; break;
	  
        case I_fcmovne:   taken = AINS_A_EVAL(iins) == 0.0 ? 0 : 1; break;
	  
        case I_fcmovlt:   taken = AINS_A_EVAL(iins) <  0.0 ? 1 : 0; break;
	  
        case I_fcmovge:   taken = AINS_A_EVAL(iins) <  0.0 ? 0 : 1; break;
	  
        case I_fcmovgt:   taken = AINS_A_EVAL(iins) >  0.0 ? 1 : 0; break;
	  
        case I_fcmovle:   taken = AINS_A_EVAL(iins) >  0.0 ? 0 : 1; break;
	  
        default:
	  taken = 1;
	  ERROR(("Unexpected FCM instruction\n"));
	  PrintIns(stdout,iins);
	  break;
        }
      
      if( taken )
        {
	  AinsMakeMove(iins,AINS_REG_B(iins),DT_TFLT,AINS_REG_C(iins),DT_TFLT);
	  return 1;
        }
      else
        {
	  AinsKill(iins);
	  return 1;
        }
      
    }
  
  if( AINS_REG_B(iins) == AINS_REG_C(iins) )
    {
      AinsKill(iins);
      return 1;
    }
  
  
  return 0;
}
#undef __PROC__
#define __PROC__ "KillBadJump"
   
static void KillBadJump(INDEX iins)
{
  INDEX jbbl;
  INDEX ibbl = AINS_BBL(iins);
  INDEX ifun = ABBL_FUN(ibbl);
  
  if( AbblNpreds(ibbl) == 0 )
    {
      AbblKill(ibbl);
      return;
    }
  
  /* NEEDS WORK: this is a bad hack but works very well */
    
  if( AINS_CODE(iins) == I_jsr )
    {
      CASSERT( ABBL_TYPE(ibbl)==BT_CALL,("jsr found in non bt_call bbl\n"));
      CASSERT( VALID(ABBL_EDG_CALL(ibbl)),("bt_call bbl with invalid edg_call\n"));
      CASSERT( VALID(ABBL_EDG_LINK(ibbl)),("bt_call bbl with invalid edg_link\n"));
      
      jbbl = AEDG_TAIL( ABBL_EDG_LINK(ibbl) );
      AinsKill(iins);
      AedgKillFancy(ABBL_EDG_CALL(ibbl));
      AedgKillFancy(ABBL_EDG_LINK(ibbl));
      ABBL_TYPE(ibbl) = BT_NORM;
      AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_NORM));                                   
    }
  else if( AINS_CODE(iins) == I_jmp )
    {
      CASSERT( ABBL_TYPE(ibbl)==BT_JUMP,("jmp found in non bt_jump bbl\n"));
      CASSERT( VALID(ABBL_EDG_JUMP(ibbl)),("bt_jump bbl with invalid edg_jump\n"));
      
      AinsKill(iins);
      AedgKillFancy(ABBL_EDG_JUMP(ibbl));        
      ABBL_TYPE(ibbl) = BT_NORM;
      AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_LAST(ifun),ET_NORM));                                   
    }
  else
    {
      WARNING(("Unexpected Jump Instruction %s (target: 0x%lx)\n",AINS_NAME(iins),AINS_B_EVAL(iins)));
      PrintBblSuper(stdout,ibbl);
    }
}
#undef __PROC__
#define __PROC__ "StrRedJMP"

static NUMBER StrRedJMP(INDEX iins)
{
  INDEX ibbl;
  INDEX isec;
  INDEX target;
  INDEX iedg;
  INDEX repl_edg;

  if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
    return 0;
  
  ibbl = AINS_BBL(iins);
  isec = LocateSection(AINS_B_EVAL(iins));
  if( isec != SysGetTextSection() )
    {
      KillBadJump(iins);
      return 0;
    }
  
  /* target iins might be killed because it is a useless
     instrucion but its bbl should be live */
    
  
  switch( AINS_CODE(iins) )
    {
    case I_jsr:
      CASSERT( ABBL_TYPE(ibbl)==BT_CALL,("jsr found in non bt_call bbl\n"));
      CASSERT( VALID(ABBL_EDG_CALL(ibbl)),("bt_call bbl with invalid edg_call\n"));        
      target = FindTargetBbl(AINS_B_EVAL(iins));
      AinsMakeBsr(iins,AINS_REG_C(iins));
      AINS_C_EVAL(iins) = Ins2Addr(iins+1);    /* NEEDS WORK */
      AINS_FLAGS(iins) |= FI_EVAL_C;

      repl_edg = ABBL_EDG_CALL(ibbl);
      AedgUnlinkFancy(repl_edg);
      AEDG_TAIL(repl_edg)=target;
      AedgLinkFancy(repl_edg);

      return 1;
      
    case I_jmp:
      target = FindTargetBbl(AINS_B_EVAL(iins));
      if( ABBL_TYPE(ibbl) != BT_JUMP )
        {
	  CASSERT( ABBL_IS_RELOC(target),("weird\n"));
	  HURRAH(("Resolving jmp in bbl: %d type %c\n",ibbl,ABBL_TYPE(ibbl)));
        }
      
      AinsKill(iins);
      FOREACH_BBL_SUCC(ibbl,iedg)
        {
	  AedgKillFancy(iedg);
        }
      
      ABBL_TYPE(ibbl) = BT_NORM;
      AedgLinkFancy( AedgNewFancy(ibbl,target,ET_NORM));
      return 1;
      
    case I_ret:
      return 0;
      
    default:
      ERROR(("Did not evaluate jsr_co\n"));
      return 0;
    }
}
#undef __PROC__
#define __PROC__ "StrRedIBR"
static NUMBER StrRedIBR(INDEX iins)  
{
  INDEX ibbl = AINS_BBL(iins);
  INDEX iedg;
  BOOL taken;

  CASSERT( ABBL_TYPE(ibbl) == BT_BRANCH,("ibr found in non bt_branch bbl\n"));
  CASSERT( VALID(ABBL_EDG_TRUE(ibbl)),("bt_branch bbl with invalid edg_true\n"));
  CASSERT( VALID(ABBL_EDG_FALSE(ibbl)),("bt_branch bbl with invalid edg_false\n"));
  
  if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
    return 0;    
  
  switch(AINS_CODE(iins))
    {
    case I_beq:   taken =  AINS_A_EVAL(iins) == 0 ? 1 : 0; break;
      
    case I_bne:   taken =  AINS_A_EVAL(iins) == 0 ? 0 : 1; break;
      
    case I_blt:   taken =  AINS_A_EVAL(iins) <  0 ? 1 : 0; break;
      
    case I_bge:   taken =  AINS_A_EVAL(iins) <  0 ? 0 : 1; break;
      
    case I_bgt:   taken =  AINS_A_EVAL(iins) >  0 ? 1 : 0; break;
        
    case I_ble:   taken =  AINS_A_EVAL(iins) >  0 ? 0 : 1; break;
      
    case I_blbs:  taken =  AINS_A_EVAL(iins) &  1 ? 1 : 0; break;
      
    case I_blbc:  taken =  AINS_A_EVAL(iins) &  1 ? 0 : 1; break;
      
    default:      taken = 1; FATAL(("weird\n"));
    }
  
  if( taken )
    {
      iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_TRUE(ibbl)),ET_NORM);
    }
  else
    {
      iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_FALSE(ibbl)),ET_NORM);
    }
  
  AedgKillFancy(ABBL_EDG_TRUE(ibbl));
  AedgKillFancy(ABBL_EDG_FALSE(ibbl));
  AinsKill(iins);
  ABBL_TYPE(ibbl) = BT_NORM;
  AedgLinkFancy(iedg);
  return 1;
}
#undef __PROC__
#define __PROC__ "StrRedFBR"
static NUMBER StrRedFBR(INDEX iins)  
{
  INDEX ibbl = AINS_BBL(iins);
  INDEX iedg;
  BOOL taken;
  
  CASSERT( ABBL_TYPE(ibbl) == BT_BRANCH,("fbr found in non bt_branch bbl\n"));
  CASSERT( VALID(ABBL_EDG_TRUE(ibbl)),("bt_branch bbl with invalid edg_true\n"));
  CASSERT( VALID(ABBL_EDG_FALSE(ibbl)),("bt_branch bbl with invalid edg_false\n"));
  
  
  if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
    return 0;    
  
  switch(AINS_CODE(iins))
    {
    case I_fbeq:   taken = AINS_A_FEVAL(iins) == 0.0 ? 1 : 0; break;
      
    case I_fbne:   taken = AINS_A_FEVAL(iins) == 0.0 ? 0 : 1; break;
      
    case I_fblt:   taken = AINS_A_FEVAL(iins) <  0.0 ? 1 : 0; break;
      
    case I_fbge:   taken = AINS_A_FEVAL(iins) <  0.0 ? 0 : 1; break;
      
    case I_fbgt:   taken = AINS_A_FEVAL(iins) >  0.0 ? 1 : 0; break;
      
    case I_fble:   taken = AINS_A_FEVAL(iins) >  0.0 ? 0 : 1; break;
      
    default:      taken = 1; FATAL(("weird\n"));
    }
  
  if( taken )
    {
      iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_TRUE(ibbl)),ET_NORM);
    }
  else
    {
      iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_FALSE(ibbl)),ET_NORM);
    }
  
  AedgKillFancy(ABBL_EDG_TRUE(ibbl));
  AedgKillFancy(ABBL_EDG_FALSE(ibbl));
  AinsKill(iins);
  ABBL_TYPE(ibbl) = BT_NORM;
  AedgLinkFancy(iedg);
  return 1;
}

/*=======================================================
  Do trivial constant detection
  =======================================================*/

static void AinsTrivialEvalNew(INDEX iins)
{
  REG reg = AINS_REG_A(iins);
  
  if( reg == REG_LIT )
    {
      AINS_A_EVAL(iins) = AINS_DATA(iins);                    
      AINS_FLAGS(iins) |= FI_EVAL_A;
    }
  else if( reg == REG_ZERO || reg == REG_FZERO )
    {
      AINS_A_EVAL(iins) = 0;
      AINS_FLAGS(iins) |= FI_EVAL_A;
    }
  
  reg = AINS_REG_B(iins);
  
  if( reg == REG_LIT )
    {
      AINS_B_EVAL(iins) = AINS_DATA(iins);                    
      AINS_FLAGS(iins) |= FI_EVAL_B;
    }
  else if( reg == REG_ZERO || reg == REG_FZERO )
    {
      AINS_B_EVAL(iins) = 0;
      AINS_FLAGS(iins) |= FI_EVAL_B;
    }
  else if( reg == REG_GP )
    {
      {
	AINS_B_EVAL(iins) = SysGetGlobalPointer();
	AINS_FLAGS(iins) |= FI_EVAL_B;
      }
    }
  
  reg = AINS_REG_C(iins);  /* important for jmp and bra evaluation */
  
  if( reg == REG_ZERO || reg == REG_FZERO )
    {
      AINS_C_EVAL(iins) = 0;
      AINS_FLAGS(iins) |= FI_EVAL_C;
    }
  
  if (AINS_TYPE(iins) == IT_JMP || AINS_TYPE(iins) == IT_BRA)
    {
      if ( !(AINS_FLAGS(iins) & FI_EVAL_C))
	{
	  AINS_C_EVAL(iins) = Ins2Addr(iins+1);   // NEEDS WORK 
	  AINS_FLAGS(iins) |= FI_EVAL_C;
	}
    }
  
}

/*==================================================================
  Evaluate a floating point load
  ===================================================================*/

static void EvalFLD(INDEX iins, VALUESPEC* spec)
{ 
  INDEX isec;
  INT64 address;

  INT64 top;
  INT64 bot;
  INT64 regbitb=REG2BIT(AINS_REG_B(iins));
  INT64 regbitc=REG2BIT(AINS_REG_C(iins));

  if (AINS_REG_B(iins)==REG_GP)
    {
      address = SysGetGlobalPointer();
      top = 0;
      bot = 0;
    }
  else
    {
      address = spec->value[AINS_REG_B(iins)].i;
      top = spec->top & regbitb;
      bot = spec->bot & regbitb;
    }
  
  if (bot | top)
    {
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
  else 
    {
      address += AINS_DATA(iins);
      isec = LocateSection(address);
      if (isec<0)
	{
	  spec->top |= regbitc;
	  spec->bot &= ~regbitc;
	  return;
	}

      if( AsecIsReadOnly(isec)
#ifdef ALTO
	  || AsecConstSectionContainsAddress(address)
#endif
	  )
	{
	  if(!AsecIsCode(isec) )
	    {
	      if (ASEC_IS_BSS(isec) || ASEC_IS_SBSS(isec))
		spec->value[AINS_REG_C(iins)].f=0.0;
	      else
		spec->value[AINS_REG_C(iins)].f= EmulFLD(AINS_CODE(iins),AsecMapAddress(isec,address));
	      AdaptMapOTF(CONST_LOAD,address,AINS_BBL(iins)); // a load of possibly non-constant data was performed
	      spec->top &= ~regbitc;
	      spec->bot &= ~regbitc;
	      return;
	    }
	}
      else if (AddressIsConstantOTF(address))
	{
	  if(!AsecIsCode(isec) )
	    {
	      if (ASEC_IS_BSS(isec) || ASEC_IS_SBSS(isec))
		spec->value[AINS_REG_C(iins)].f=0.0;
	      else
		spec->value[AINS_REG_C(iins)].f= EmulFLD(AINS_CODE(iins),AsecMapAddress(isec,address));
	      AdaptMapOTF(CONST_LOAD,address,AINS_BBL(iins)); // A load from possibly non-constant data was performed
	      spec->top &= ~regbitc;
	      spec->bot &= ~regbitc;
	      return;
	    }
	}
      else
	if(!AsecIsCode(isec) )
	  AdaptMapOTF(CONST_LOAD,address,AINS_BBL(iins));
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
}  
/*==================================================================
  Evaluate integer instructions
  ===================================================================*/

static void EvalIOP(INDEX iins, VALUESPEC* spec)
{
  INT64 int1,int2,int3 = 0;
  INDEX sectext =  SysGetTextSection();
  REG rega,regb;

  INT64 bot;
  INT64 top;
  INT64 regbita = REG2BIT(AINS_REG_A(iins));
  INT64 regbitb = REG2BIT(AINS_REG_B(iins));
  INT64 regbitc = REG2BIT(AINS_REG_C(iins));

  rega = AINS_REG_A(iins);
  if (rega==REG_ZERO || rega==REG_FZERO || rega == REG_LIT)
    {
      int1 = AINS_A_EVAL(iins);
      top = 0;
      bot = 0;
    }
  else 
    {
      int1 = spec->value[AINS_REG_A(iins)].i;
      top = spec->top & regbita;
      bot = spec->bot & regbita;
    }
  
  regb = AINS_REG_B(iins);
  if (regb==REG_ZERO || regb==REG_FZERO || regb == REG_LIT)
    int2 = AINS_B_EVAL(iins);
  else if (regb==REG_GP)
    int2 = SysGetGlobalPointer();
  else 
    {
      int2 = spec->value[AINS_REG_B(iins)].i;
      top |= spec->top & regbitb;
      bot |= spec->bot & regbitb;
    }
    
  if ((bot | top) ||  LocateSection(int1) == sectext || LocateSection(int2) == sectext)
    {
      if (!((spec->top & regbita) || (spec->bot & regbita)))
	{
	  AdaptMapOTF(BLOCK_STORE,int1,AINS_BBL(iins));
	}
      if (!((spec->top & regbitb) || (spec->bot & regbitb)))
	{
	  AdaptMapOTF(BLOCK_STORE,int2,AINS_BBL(iins));
	}
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
  else
    {
      if (AINS_TYPE(iins) == IT_ICM)
	{
	  switch (AINS_CODE(iins))
	    {
	    case I_cmovlbc:
	      if (!(int1 & 1L))
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmoveq:
	      if (int1 == 0L)
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmovlt:
	      if (int1 < 0L)
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmovle:
	      if (int1 <= 0L)
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmovlbs:
	      if (int1 & 1L)
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmovne:
	      if (int1 != 0L)
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmovge:
	      if (int1 >= 0L)
		int3 = int2;
	      else
		return;
	      break;
	    case I_cmovgt:
	      if (int1 > 0L)
		int3 = int2;
	      else
		return;
	      break;
	    }
	}
      else
	{
	  int3 = EmulIOP(AINS_CODE(iins),int1,int2);
	}
      spec->value[AINS_REG_C(iins)].i = int3;
      spec->top &= ~regbitc;
      spec->bot &= ~regbitc;
    }
}

/*==================================================================
  evaluate integer loads
  ===================================================================*/
static void EvalILD(INDEX iins, VALUESPEC* spec)
{
  INDEX isec;
  INT64 address;
  
  INT64 top;
  INT64 bot;
  INT64 regbitb=REG2BIT(AINS_REG_B(iins));
  INT64 regbitc=REG2BIT(AINS_REG_C(iins));
  
  if (AINS_REG_B(iins)==REG_GP) 
    {
      address = SysGetGlobalPointer();
      top = 0;
      bot = 0;
    }
  else
    {
      address = spec->value[AINS_REG_B(iins)].i;
      top = spec->top & regbitb;
      bot = spec->bot & regbitb;
    }

  if (bot | top)
    {
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
  else 
    {
      address += AINS_DATA(iins);
      isec = LocateSection(address);
      
      if (isec<0)
	{
	  spec->top |= regbitc;
	  spec->bot &= ~regbitc;
	  return;
	}

      if( AsecIsReadOnly(isec) ||
#ifdef ALTO
	  AsecConstSectionContainsAddress(address) ||
#endif
	  AsecSwitchSectionContainsAddress(address)
	  )
	{
	  if(!AsecIsCode(isec) )
	    {
	      if (ASEC_IS_BSS(isec) || ASEC_IS_SBSS(isec))
		{
		  spec->value[AINS_REG_C(iins)].i=0L;
		}
	      else
		{
		  spec->value[AINS_REG_C(iins)].i= EmulILD(AINS_CODE(iins),AsecMapAddress(isec,address));
		}
	      AdaptMapOTF(CONST_LOAD,address,AINS_BBL(iins)); // a load of possibly non-constant data was performed
	      spec->top &= ~regbitc;
	      spec->bot &= ~regbitc;
	      return;
	    }
	}
      else if (AddressIsConstantOTF(address))
	{
	  if(!AsecIsCode(isec) )
	    {
	      if (ASEC_IS_BSS(isec) || ASEC_IS_SBSS(isec))
		spec->value[AINS_REG_C(iins)].i=0L;
	      else
		spec->value[AINS_REG_C(iins)].i= EmulILD(AINS_CODE(iins),AsecMapAddress(isec,address));
	      AdaptMapOTF(CONST_LOAD,address,AINS_BBL(iins)); // a load of possibly non-constant data was performed
	      spec->top &= ~regbitc;
	      spec->bot &= ~regbitc;
	      return;
	    }
	}
      else
	if(!AsecIsCode(isec) )
	  AdaptMapOTF(CONST_LOAD,address,AINS_BBL(iins));
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
}

static void EvalLDA(INDEX iins, VALUESPEC* spec)
{
  INT64 address;

  INT64 top;
  INT64 bot;
  INT64 regbitb=REG2BIT(AINS_REG_B(iins));
  INT64 regbitc=REG2BIT(AINS_REG_C(iins));

  if (AINS_REG_B(iins)==REG_GP) 
    {
      address = SysGetGlobalPointer();
      top = 0;
      bot = 0;
    }
  else
    {
      address = spec->value[AINS_REG_B(iins)].i;
      top = spec->top & regbitb;
      bot = spec->bot & regbitb;
    }
  
  if (bot | top)
    {
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
  else 
    { 
      if( AINS_CODE(iins) == I_lda )
	address += AINS_DATA(iins);
      if( AINS_CODE(iins) == I_ldah )
	address += ((INT64)AINS_DATA(iins)<<16L);
      
      spec->value[AINS_REG_C(iins)].i = address;
      spec->top &= ~regbitc;
      spec->bot &= ~regbitc;
    }
}
/*==================================================================
  evaluate floating-point operations
  ===================================================================*/

static void EvalFOP(INDEX iins, VALUESPEC* spec)
{
  INT64 int1,int2;
  double float1,float2,float3;

  INT64 bot;
  INT64 top;
  INT64 regbita = REG2BIT(AINS_REG_A(iins));
  INT64 regbitb = REG2BIT(AINS_REG_B(iins));
  INT64 regbitc = REG2BIT(AINS_REG_C(iins));

  float3 = 0.0;

  top = spec->top & regbita;
  bot = spec->bot & regbita;

  top |= spec->top & regbitb;
  bot |= spec->bot & regbitb;
  
  if (top | bot)
    {
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }
  else
    {
      int1 = spec->value[AINS_REG_A(iins)].i;
      float1 = spec->value[AINS_REG_A(iins)].f;
      
      int2 = spec->value[AINS_REG_B(iins)].i;
      float2 = spec->value[AINS_REG_B(iins)].f;


      if ( !(int1 & DOUBLE_INF) || !(int2 & DOUBLE_INF) || int1==DOUBLE_INF || int2==DOUBLE_INF)
	{
	  spec->top |= regbitc;
	  spec->bot &= ~regbitc;
	}
      else
	{
	  if (AINS_TYPE(iins)==IT_FCM)
	    {
	      switch (AINS_CODE(iins))
		{
		case I_fcmoveq:
		  if (float1 == 0.0)
		    float3 = float2;
		  else
		    return;
		  break;
		case I_fcmovne:
		  if (float1 != 0.0)
		    float3 = float2;
		  else
		    return;
		  break;
		case I_fcmovlt:
		  if (float1 < 0.0)
		    float3 = float2;
		  else
		    return;
		  break;
		case I_fcmovge:
		  if (float1 >= 0.0)
		    float3 = float2;
		  else
		    return;
		  break;
		case I_fcmovgt:
		  if (float1 > 0.0)
		    float3 = float2;
		  else
		    return;
		  break;
		case I_fcmovle:  
		  if (float1 <= 0.0)
		    float3 = float2;
		  else
		    return;
		}
	    }
	  else
	    {
	      float3=EmulFOP(AINS_CODE(iins),float1,float2);
	    }
	  spec->value[AINS_REG_C(iins)].f = float3;
	  spec->top &= ~regbitc;
	  spec->bot &= ~regbitc;  
	}
    }    
  
  return;
}

/*==================================================================
  evaluate stores: data might become non-constant
  ===================================================================*/

void EvalST(INDEX iins, VALUESPEC* spec)
{
  INT64 address;
  
  INT64 top;
  INT64 bot;
  INT64 regbitb=REG2BIT(AINS_REG_B(iins));
  INT64 regbita=REG2BIT(AINS_REG_A(iins));

  if (AINS_REG_B(iins)==REG_GP) 
    {
      address = SysGetGlobalPointer();
      top = 0;
      bot = 0;
    }
  else
    {
      address = spec->value[AINS_REG_B(iins)].i;
      top = spec->top & regbitb;
      bot = spec->bot & regbitb;
    }
 
  if (!(bot | top))
    {
      address += AINS_DATA(iins);

      if (AddressIsConstantOTF(address))
	{
	  /* data is stored at a location */
	  AdaptMapOTF(CONST_STORE,address,AINS_BBL(iins));
	}
    }
  
  address = spec->value[AINS_REG_A(iins)].i;
  top = spec->top & regbita;
  bot = spec->bot & regbita;
  
  if (!(bot | top))
    {
      /* an address is stored somewhere */
      AdaptMapOTF(BLOCK_STORE,address,AINS_BBL(iins));
#if DETECT_DEAD_CALL_EDGES
      if (AddressIsCode(address))
	{
	  // set edge pointed to by address at location live
	  INDEX ifun=ABBL_FUN(AINS_BBL(Addr2Ins(address)));
	  INDEX bbl=AFUN_BBL_FIRST(ifun);
	  INDEX iedg;
	  AFUN_FLAGS(ifun) |= F_MARKED;
	  FOREACH_BBL_PRED(bbl,iedg)
	    if (ABBL_FUN(AEDG_HEAD(iedg))==AFUN_PSEUDO_HELL_LITE)
	      {
		AEDG_MARK(iedg);
		if (!AEDG_LIVE(iedg))
		  {
		    super_change=1;
		    //		    printf("1 making edge to %d %s live\n",ifun,AFUN_NAME(ifun));
		  }
		AEDG_SET_LIVE(iedg);
	      }
	}
#endif
    }
}

void EvalIFC(INDEX iins, VALUESPEC* spec)
{
  INT64 regbita=REG2BIT(AINS_REG_A(iins));
  INT64 regbitb=REG2BIT(AINS_REG_B(iins));
  INT64 regbitc=REG2BIT(AINS_REG_C(iins));

  spec->value[AINS_REG_C(iins)].i=spec->value[AINS_REG_A(iins)].i;
  spec->bot &= ~regbitc;
  spec->top &= ~regbitc;

  if (spec->bot & regbita)
    {
      spec->bot |= regbitc;
      spec->top &= ~regbitc;
    }
  else if (spec->top & regbita)
    {
      spec->top |= regbitc;
      spec->bot &= ~regbitc;
    }

}

/*==================================================================
  evaluate a bbl for incoming register values
  ===================================================================*/

static void AbblEvalValueSpec(INDEX ibbl, VALUESPEC* output)
{
  INDEX iins;
  VALUESPEC * input = GetBblValueSpec(ibbl);
  
  CopyValueSpec(input,output);

  FOREACH_BBL_INS(ibbl,iins)
    {
      if (AINS_FLAGS(iins) & F_KILLED) continue;
      AinsEvalValueSpecNew(iins,output);
    }
}

static void IntraEdgEvalValueSpec(INDEX iedg, VALUESPEC* output)
{
  INDEX iins;
  INDEX ibbl=AEDG_TAIL(iedg);
  VALUESPEC * input = GetEdgValueSpec(iedg);
  
  if (AEDG_TYPE(iedg)==ET_FUN_LINK)
    JoinValueSpec(GetEdgValueSpec(ABBL_EDG_CALL(AEDG_HEAD(iedg))+1),
		  input,
		  AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg))))),
		  AbblRegsLiveIn(AEDG_TAIL(iedg))
		  );

  CopyValueSpec(input,output);
  
  FOREACH_BBL_INS(ibbl,iins)
    {
      if (AINS_FLAGS(iins) & F_KILLED) continue;
      AinsEvalValueSpecNew(iins,output);
    }
}

/*==================================================================
  evaluate instructions giving the incoming register values
  ===================================================================*/

extern void AinsEvalValueSpecNew(INDEX iins, VALUESPEC* spec)
{
  if (AINS_REG_C(iins) == REG_ZERO || AINS_REG_C(iins) == REG_FZERO)
    if (!AINS_IS_STORE(iins))
      return;
  
  switch( AINS_TYPE(iins) )
    {
    case IT_LLD:
      if (AINS_REG_C(iins)!=31)
	{
	  spec->top |= REG2BIT(AINS_REG_C(iins));
	  spec->bot &= ~REG2BIT(AINS_REG_C(iins));;
	}
      break;
    case IT_ULD:
    case IT_ILD:
      EvalILD(iins,spec);
      break;
    case IT_IST:
    case IT_UST:
    case IT_FST:
    case IT_CST:
      EvalST(iins,spec);
      break;
    case IT_JMP:
    case IT_BRA:
      if (AINS_FLAGS(iins) & FI_EVAL_C)
	{
	  spec->value[AINS_REG_C(iins)].i = AINS_C_EVAL(iins);
	  spec->top &= ~REG2BIT(AINS_REG_C(iins));
	  spec->bot &= ~REG2BIT(AINS_REG_C(iins));
	}
      break;
    case IT_LDA:
      EvalLDA(iins,spec);
      break;
    case IT_FCV:
      spec->top |= REG2BIT(AINS_REG_C(iins));
      spec->bot &= ~REG2BIT(AINS_REG_C(iins));
      break;
    case IT_FCM:
    case IT_FCP:
    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
    case IT_FA4:
    case IT_FA8:
    case IT_FS4:
    case IT_FS8:
    case IT_FM4:
    case IT_FM8:
    case IT_SQR:
      EvalFOP(iins,spec);
      break;
    case IT_FD4:
    case IT_FD8:
      if ( !( (spec->bot|spec->top) & REG2BIT(AINS_REG_B(iins)) ) && spec->value[AINS_REG_B(iins)].f == 0.0)
	{
	  spec->top |= REG2BIT(AINS_REG_B(iins));
	  spec->bot &= ~REG2BIT(AINS_REG_B(iins));
	  spec->top |= REG2BIT(AINS_REG_C(iins));
	  spec->bot &= ~REG2BIT(AINS_REG_C(iins));
	  break;
	}
      EvalFOP(iins,spec);
      break;
    case IT_EXT:
    case IT_ICM:
    case IT_IOP:
    case IT_ICP:
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
    case IT_LOG:
    case IT_SFT:     
      EvalIOP(iins,spec);
      break;
    case IT_FLD:
      EvalFLD(iins,spec);
      break;
    case IT_MSC:
      if (!strcmp(CONFIG_MACHINE,"ev67"))
	{
	  if (AINS_CODE(iins)==I_implver)
	    {
	      spec->top &= ~REG2BIT(AINS_REG_C(iins));
	      spec->bot &= ~REG2BIT(AINS_REG_C(iins));
	      spec->value[AINS_REG_C(iins)].i=2L;
	      break;
	    }
	  if (AINS_CODE(iins)==I_amask)
	    {
	      spec->top &= ~REG2BIT(AINS_REG_C(iins));
	      spec->bot &= ~REG2BIT(AINS_REG_C(iins));
	      spec->value[AINS_REG_C(iins)].i=AINS_DATA(iins)&0x307L;
	      break;
	    }
	}
      
      spec->top |= REG2BIT(AINS_REG_C(iins));
      spec->bot &= ~REG2BIT(AINS_REG_C(iins));
      break;
    case IT_I2F:
    case IT_F2I:
      EvalIFC(iins,spec);
    default:
      break;
    }

  if (AINS_FLAGS(iins) & FI_EVAL_C)
    {
      spec->value[AINS_REG_C(iins)].i = AINS_C_EVAL(iins);
      spec->top &= ~REG2BIT(AINS_REG_C(iins));
      spec->bot &= ~REG2BIT(AINS_REG_C(iins));
    }

  return;
}

/*==================================================================
  kill edges in a fun that were not set live by the conditional
  constant propagation
  ===================================================================*/

static void AfunKillEdges(INDEX ifun)
{
  INDEX ibbl;
  INDEX last_iins;
  INDEX new_edge;
  INDEX succ,succ_edge;
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (ABBL_FLAGS(ibbl) & F_KILLED) continue;
      if (ABBL_TYPE(ibbl)!=BT_BRANCH) continue;
      if (AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)) && AEDG_IS_MARKED(ABBL_EDG_FALSE(ibbl))) continue;
      if (!AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)) && !AEDG_IS_MARKED(ABBL_EDG_FALSE(ibbl))) continue;
      
      if (AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)))
	{
	  succ_edge = ABBL_EDG_TRUE(ibbl);
	}
      else
	{
	  succ_edge = ABBL_EDG_FALSE(ibbl);
	}
      succ = AEDG_TAIL(succ_edge);
      new_edge = AedgNewFancy(ibbl,succ,ET_NORM);
      AEDG_COUNT(new_edge)=AEDG_COUNT(succ_edge);
      AedgKillFancy(ABBL_EDG_TRUE(ibbl));
      AedgKillFancy(ABBL_EDG_FALSE(ibbl));
      last_iins=INVALID;
      FOREACH_BBL_INS_R(ibbl,last_iins)
	if (AINS_TYPE(last_iins) == IT_IBR || AINS_TYPE(last_iins) == IT_FBR) break;
      AinsKill(last_iins);
      ABBL_TYPE(ibbl) = BT_NORM;
      AedgLinkFancy(new_edge);
    }
}

static void AfunKillEdgesIntraEdg(INDEX ifun)
{
  INDEX ibbl;
  INDEX last_iins;
  INDEX new_edge;
  INDEX succ,succ_edge;
  INT32 spec;

  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (ABBL_FLAGS(ibbl) & F_KILLED) continue;
      if (ABBL_TYPE(ibbl)!=BT_BRANCH) continue;
      if (AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)) && AEDG_IS_MARKED(ABBL_EDG_FALSE(ibbl))) continue;
      if (!AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)) && !AEDG_IS_MARKED(ABBL_EDG_FALSE(ibbl))) continue;
      
      if (AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)))
	{
	  succ_edge = ABBL_EDG_TRUE(ibbl);
	}
      else
	{
	  succ_edge = ABBL_EDG_FALSE(ibbl);
	}
      spec=AEDG_VALUESPEC(succ_edge);
      succ = AEDG_TAIL(succ_edge);
      new_edge = AedgNewFancy(ibbl,succ,ET_NORM);
      AEDG_COUNT(new_edge)=AEDG_COUNT(succ_edge);
      AedgKillFancy(ABBL_EDG_TRUE(ibbl));
      AedgKillFancy(ABBL_EDG_FALSE(ibbl));
      last_iins=INVALID;
      FOREACH_BBL_INS_R(ibbl,last_iins)
	if (AINS_TYPE(last_iins) == IT_IBR || AINS_TYPE(last_iins) == IT_FBR) break;
      AinsKill(last_iins);
      ABBL_TYPE(ibbl) = BT_NORM;
      AedgLinkFancy(new_edge);
      AEDG_VALUESPEC(new_edge)=spec;
    }
}

/*==================================================================
  see whether edges will be removed after ccp
  ===================================================================*/
static INT64 AfunTestKillEdges(INDEX ifun)
{
  INDEX ibbl;
  INT64 count = 0;
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (ABBL_FLAGS(ibbl) & F_KILLED) continue;
      if (ABBL_TYPE(ibbl)!=BT_BRANCH) continue;
      if (AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)) && AEDG_IS_MARKED(ABBL_EDG_FALSE(ibbl))) continue;
      if (!AEDG_IS_MARKED(ABBL_EDG_TRUE(ibbl)) && !AEDG_IS_MARKED(ABBL_EDG_FALSE(ibbl))) continue;
      if (!count) count=1;
      count+=ABBL_COUNT(ibbl);
      
    }
  return count;
}
/*==================================================================
  Before starting with the CP, convert the trivial cases of
  jsrs into bsrs
  ===================================================================*/

static void PreliminaryJsrReduction()
{
  INDEX ibbl;
  INDEX iins;
  VALUE Top;
  VALUESPEC tmp1;
  Top.i = TheLatticeTop;
  
  FOREACH_LIVE_BBL(ibbl)
    {
      if (ABBL_TYPE(ibbl)==BT_CALL)
	{
	  if (AINS_CODE(ABBL_INS_LAST(ibbl))==I_jsr && AINS_REG_B(ABBL_INS_LAST(ibbl))!=AINS_REG_C(ABBL_INS_LAST(ibbl)))
	    {
	      REG sreg;
	      INDEX jins;
	      VALUESPEC * output=&tmp1;//(VALUESPEC *) malloc(sizeof(VALUESPEC));
	      sreg=AINS_REG_B(iins=ABBL_INS_LAST(ibbl));
	      
	      FOREACH_BBL_INS(ibbl,jins)
		{
		  if (AINS_FLAGS(iins) & F_KILLED) continue;
		  AinsTrivialEvalNew(jins);
		}
	      NewBblValueSpec(ibbl,Top);
	      AbblEvalValueSpec(ibbl, output);
	      if (!((output->top | output->bot) & REG2BIT(sreg)))
		{
		  AINS_FLAGS(iins) |= FI_EVAL_B;
		  AINS_B_EVAL(iins) = output->value[sreg].i;
		  StrRedJMP(iins);
		}
	      //	      free(output);
	    }
	}
    }
}
 
#if LATE_CALL_JOIN
/*==================================================================
  propagate each calling context into the procedures,
  calculate the meet at the instructions
  ===================================================================*/
static void ContextSensitiveElimination()
{
  INDEX ifun;
  INDEX ibbl;
  INDEX iins;
  BOOL flag;
  INDEX iedg;
  VALUE Bot;
  Bot.i = TheLatticeBot;

  FOREACH_LIVE_INS(iins)
    {
      if (!(AINS_FLAGS(iins) & FI_EVAL_A))
	AINS_A_EVAL(iins)=TheLatticeBot;
      if (!(AINS_FLAGS(iins) & FI_EVAL_B))
	AINS_B_EVAL(iins)=TheLatticeBot;
      if (!(AINS_FLAGS(iins) & FI_EVAL_C))
	AINS_C_EVAL(iins)=TheLatticeBot;
    }
  
  FOREACH_LIVE_FUN_R(ifun)
    {
      INDEX call;
    
      if (AFUN_IS_PSEUDO(ifun)) continue;
      
      FOREACH_FUN_BBL(ifun,ibbl)
	if (ABBL_TYPE(ibbl)==BT_BRANCH)
	  {
	    AEDG_UNMARK(ABBL_EDG_TRUE(ibbl));
	    AEDG_UNMARK(ABBL_EDG_FALSE(ibbl));
	  }
      
      FOREACH_FUN_BBL(ifun,ibbl)
	FOREACH_BBL_INS(ibbl,iins)
	  AINS_FLAGS(iins)&=~F_MARKED;
      
      flag = 0;

      FOREACH_BBL_PRED(AFUN_BBL_FIRST(ifun),call)
	{
	  if (!AEDG_LIVE(call)) continue; // OPPASSEN
	  if (!AEDG_IS_INTER(call)) continue;
	  flag = 1;
	  if (0 && ifun>CONFIG_MIN && ifun<CONFIG_MAX)
	    {
	      AfunIntraEdgSetCI(ifun);
	      AfunConstPropIntraEdg(ifun,call);
	    }
	  else
	    {
	      AfunSetCS(ifun,call);
	      AfunConstProp(ifun);
	    }

	  FOREACH_FUN_BBL(ifun,ibbl)
	    if (ABBL_LIVE(ibbl))
	      {
		if (0 && ifun>CONFIG_MIN && ifun<CONFIG_MAX)
		  {
		    NewBblValueSpec(ibbl,Bot);
		    FOREACH_BBL_PRED(ibbl,iedg)
		      {
			if (!(AEDG_LIVE(iedg))) continue;
			if (AEDG_TYPE(iedg)==ET_FUN_RETURN) continue;
			CopyValueSpec(GetEdgValueSpec(iedg),GetBblValueSpec(ibbl));
			if (AEDG_TYPE(iedg)==ET_FUN_LINK)
			  JoinValueSpec(GetEdgValueSpec(ABBL_EDG_CALL(AEDG_HEAD(iedg))+1),
					GetBblValueSpec(ibbl),
					AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg))))),
					AbblRegsLiveIn(AEDG_TAIL(iedg))
					);
			/* mark non-idempotent instructions */
			AbblMarkValueSpec(ibbl);
			/* set tmp register values in instructions */
			AbblProvSetValueSpec(ibbl);
		      }
		  }
		else
		  {
		    /* mark non-idempotent instructions */
		    AbblMarkValueSpec(ibbl);
		    /* set tmp register values in instructions */
		    AbblProvSetValueSpec(ibbl);
		  }
	      }
	  
	  ClearIntraEdgValueSpecsFun(ifun);
	  ClearBblValueSpecsFun(ifun);
	} 
      
      if (flag)
	{

	  FOREACH_FUN_BBL(ifun,ibbl)
	    {
	      if (!(ABBL_FLAGS(ibbl) & F_KILLED))
		if (ABBL_LIVE(ibbl))
		  FOREACH_BBL_INS(ibbl,iins)
		{
#if ELIMINATION
		  /* kill idempotent instructions */
		  if (!(AINS_FLAGS(iins) & F_KILLED))
		    if (!(AINS_FLAGS(iins) & F_MARKED))
		      if (!(AINS_FLAGS(iins)&FI_EVAL_C))
			AinsKill(iins);
#endif	      
		  /* set flags */
		  if (!(AINS_FLAGS(iins) & F_KILLED))
		    {
		      if (!(AINS_FLAGS(iins) & FI_EVAL_A))
			if (AINS_A_EVAL(iins) != TheLatticeBot && AINS_A_EVAL(iins) != TheLatticeTop)
			  AINS_FLAGS(iins) |= FI_EVAL_A;
		      if (!(AINS_FLAGS(iins) & FI_EVAL_B))
			if (AINS_B_EVAL(iins) != TheLatticeBot && AINS_B_EVAL(iins) != TheLatticeTop)
			  AINS_FLAGS(iins) |= FI_EVAL_B;
		      if (!(AINS_FLAGS(iins) & FI_EVAL_C))
			if (AINS_C_EVAL(iins) != TheLatticeBot && AINS_C_EVAL(iins) != TheLatticeTop)
			  AINS_FLAGS(iins) |= FI_EVAL_C;
		    }
		}
	    }
	  
	  if (AfunTestKillEdges(ifun))
	    {
	      AfunKillEdges(ifun);
	      
	      // re-assignment of the constants
	      
	      ClearBblValueSpecsFun(ifun);
	      AfunSetCI(ifun);
	      AfunConstProp(ifun);
	      FOREACH_FUN_BBL(ifun,ibbl)
		{
		  if (!ABBL_LIVE(ibbl)) continue;
		  AbblSetValueSpec(ibbl);
		}          
	      ClearBblValueSpecsFun(ifun);
	    }
	}
    }   
}
#endif

/*==================================================================
  Init OTF data analysis enabled
  ===================================================================*/

void InitBlocks4CP()
{
  int tel,tel2;
  MAP_BLOCK * tmp;

  for (tel=0; tel<numsections; tel++)
    for (tmp=map_sections[tel].first; tmp!=NULL; tmp=tmp->next)
      {
	tmp->otf_status=(char) 0x00;
	if (tmp->otf_map)
	  for (tel2=0;tel2<tmp->size;tel2+=8)
	    tmp->otf_map[tel2/8]=0;
      }
  OptDataInsScanOTF();
  Relocs2Map();
  
}
/*==================================================================
  Init OTF data analysis disabled
  ===================================================================*/

void OffInitBlocks4CP()
{
  int tel,tel2;
  MAP_BLOCK * tmp;

  for (tel=0; tel<numsections; tel++)
    for (tmp=map_sections[tel].first; tmp!=NULL; tmp=tmp->next)
      {
	tmp->otf_status=tmp->status;
	if (tmp->otf_map)
	  for (tel2=0;tel2<tmp->size;tel2+=8)
	    tmp->otf_map[tel2/8]=tmp->map[tel2/8] & ~IS_REFQUAD;
      }
  
  Relocs2Map();
}

static void AbblJoinIntraEdgValueSpecs(INDEX ibbl)
{
  INDEX iedg;
  VALUE Bot;
  Bot.i = TheLatticeBot;

  NewBblValueSpec(ibbl,Bot);

  FOREACH_BBL_PRED(ibbl,iedg)
    {
      if (AEDG_TYPE(iedg)==ET_FUN_RETURN) continue;

      JoinValueSpec(GetEdgValueSpec(iedg),GetBblValueSpec(ibbl),REGS_ALL,AbblRegsLiveIn(ibbl));
      if (AEDG_TYPE(iedg)==ET_FUN_LINK)
	JoinValueSpec(GetEdgValueSpec(ABBL_EDG_CALL(AEDG_HEAD(iedg))+1),
		      GetBblValueSpec(ibbl),
		      AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AEDG_HEAD(iedg))))),
		      AbblRegsLiveIn(AEDG_TAIL(iedg))
		      );
    }
}

#undef __PROC__
#define __PROC__ "OptPartialEvaluationNew"

/*==================================================================
  Perform a full partial evaluation
  ===================================================================*/

extern void OptPartialEvaluationNew(FILE *fp)
{
  INDEX iedg;
  INDEX ifun;
  BOOL change;
  VALUE Bot;
  VALUE Top;
  INDEX ibbl;
  INDEX iins;
  REG reg;
  clock_t time;

  Top.i = TheLatticeTop;
  Bot.i = TheLatticeBot;

  inoptdata=0;

  ComputeEdgeCounts(stdout,1,0);

  OffInitBlocks4CP();

  PreliminaryJsrReduction();

  /* no OTF data-analysis after factoring: previously found constants might not be found anymore,
     which would give problems */

  if (FACTOR_DONE || HasOption(DISABLED_OPT,"OTF"))
    OffInitBlocks4CP();
  else
    InitBlocks4CP();

  time = clock();  

  /* OTF conversion of jsrs into bsr is initialized */
  FOREACH_BBL(ibbl)
    ABBL_FLAGS(ibbl)&=~FB_REEDGED;

  ComputeSavedChangedRegisters(stdout);
  ComLiveness(stdout);

  // Unmark RA uses
  
  FOREACH_LIVE_INS(iins)
    {
      AinsTrivialEvalNew(iins);
      if (AINS_TYPE(iins)==IT_BRA || AINS_TYPE(iins)==IT_JMP) continue;
      if (AINS_REG_A(iins)==REG_RA)
	AINS_FLAGS(iins) &= ~FI_EVAL_A;
      if (AINS_REG_B(iins)==REG_RA)
	AINS_FLAGS(iins) &= ~FI_EVAL_B;
      if (AINS_REG_C(iins)==REG_RA)
	AINS_FLAGS(iins) &= ~FI_EVAL_C;
    }
      
  // Create ValueSpecs & mark edges, funs and bbls
  
  FOREACH_LIVE_FUN(ifun)
    AFUN_UNMARK(ifun);
  
  FOREACH_LIVE_EDG(iedg)
    {
      AEDG_UNLIVE(iedg);
      AEDG_UNMARK(iedg);
    }
    
  ClearBblValueSpecs();
  ClearEdgValueSpecs();
  ClearIntraEdgValueSpecs();
  
  FOREACH_LIVE_EDG(iedg)
    {
      if (AFUN_IS_PSEUDO(ABBL_FUN(AEDG_HEAD(iedg)))) continue;
      if (AFUN_IS_PSEUDO(ABBL_FUN(AEDG_TAIL(iedg)))) continue;
      if (AEDG_IS_INTER(iedg))
	{
	  if (INTERPROCEDURAL_CONSTANT_PROPAGATION)
	    NewEdgValueSpec(iedg,Bot);
	  else
	    NewEdgValueSpec(iedg,Top);
	}
    }
  
  for (ibbl=0;ibbl<4;ibbl++)
    {
      FOREACH_BBL_SUCC(ibbl,iedg)
	{
	  if (AEDG_FLAGS(iedg) & F_KILLED) continue;
	  NewEdgValueSpec(iedg,Top);
#if DETECT_DEAD_CALL_EDGES
	  if (ibbl!=AFUN_BBL_FIRST(AFUN_PSEUDO_HELL_LITE) || FACTOR_DONE || HasOption(DISABLED_OPT,"OTF"))
#endif
	    {
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(iedg)));
	      AEDG_MARK(iedg);
	      AEDG_SET_LIVE(iedg);
	    }
	}
         
      FOREACH_BBL_PRED(ibbl,iedg)
	{
	  if (AEDG_FLAGS(iedg) & F_KILLED) continue;
	  NewEdgValueSpec(iedg,Top);
	  AEDG_MARK(iedg);
	  AEDG_SET_LIVE(iedg);
	}
    }

  FOREACH_BBL_PRED(AFUN_BBL_FIRST(AFUN_PSEUDO_CALL),iedg)
    {
      REGSET_FOREACH_REG(~(REGS_IARG | REGS_FARG),reg)
	SetOneEdgValueSpec(iedg,Bot,reg);
    }
  
  FOREACH_LIVE_BBL(ibbl)
    {
      ABBL_UNLIVE(ibbl);
      if (ABBL_FLAGS(ibbl) & FB_PROG_ENTRY)
	{
	  AFUN_MARK(ABBL_FUN(ibbl));
	  FOREACH_BBL_PRED(ibbl,iedg)
	    {
	      AEDG_SET_LIVE(iedg);
	      AEDG_MARK(iedg);
	      NewEdgValueSpec(iedg,Top);
	    }
	  ABBL_SET_LIVE(ibbl);
	  ABBL_MARK(ibbl);
	}
    }
  
  // perform the first fixpoint over the whole program

#if PROPAGATION_METHOD
  change = 1;
  super_change=0;
  while (change || super_change)
    {
      change = 0;
      super_change = 0;
      FOREACH_LIVE_FUN(ifun)
	{
	  if (!AFUN_IS_MARKED(ifun)) continue;
	  if (AFUN_IS_PSEUDO(ifun)) continue;
	  AFUN_UNMARK(ifun);
	  if (0 && ifun>CONFIG_MIN && ifun<CONFIG_MAX)
	    {
	      AfunIntraEdgSetCI(ifun);
	      AfunConstPropIntraEdg(ifun,-1);
	      change |= AfunIntraEdgChangeCI(ifun);	  
	      ClearIntraEdgValueSpecsFun(ifun);
	    }
	  else
	    {
	      AfunSetCI(ifun);
	      AfunConstProp(ifun);
	      change |= AfunChangeCI(ifun);	  
	      ClearBblValueSpecsFun(ifun);
 	    }
	}
    }
#else
  change = 1;
  super_change=0;
  while (change || super_change)
    {
      change = 0;
      super_change = 0;
      FOREACH_LIVE_FUN(ifun)
	{
	  if (!AFUN_IS_MARKED(ifun)) continue;
	  if (AFUN_IS_PSEUDO(ifun)) continue;
	  AFUN_UNMARK(ifun);

	  /* before propagating all calling contexts into a procedure, the jsr-edge-conversion is initialized */

	  FOREACH_FUN_BBL(ifun,ibbl)
	    ABBL_FLAGS(ibbl)&=~FB_REEDGED;

	  FOREACH_BBL_PRED(AFUN_BBL_FIRST(ifun),iedg)
	    if ((AEDG_IS_INTER(iedg)) && AEDG_IS_MARKED(iedg))
	      {
		
		if (0 && ifun>CONFIG_MIN && ifun<CONFIG_MAX)
		  {
		    AEDG_UNMARK(iedg);
		    AfunIntraEdgSetCI(ifun);
		    AfunConstPropIntraEdg(ifun,iedg);
		    change |= AfunIntraEdgChangeCS(ifun,iedg);	  
		    ClearIntraEdgValueSpecsFun(ifun);
		  }
		else
		  {
		    AEDG_UNMARK(iedg);
		    AfunSetCS(ifun,iedg);
		    AfunConstProp(ifun);
		    change |= AfunChangeCS(ifun,iedg);	  
		    ClearBblValueSpecsFun(ifun);
		  }
	      }
	}
    }
#endif
  
  FOREACH_BBL_SUCC(AFUN_BBL_FIRST(AFUN_PSEUDO_HELL_LITE),iedg)
    {
      if (!AEDG_LIVE(iedg))
	{
	  int count = ABBL_COUNT(AEDG_TAIL(iedg));
	  INDEX pred;
	  FOREACH_BBL_PRED(AEDG_TAIL(iedg),pred)
	    if (pred!=iedg)
	      count-=AEDG_COUNT(pred);
	  //	  printf("dead edge with count %d to %d %s\n",count,ABBL_FUN(AEDG_TAIL(iedg)),AFUN_NAME(ABBL_FUN(AEDG_TAIL(iedg))));
	}
    }
  

  /* blocks that had data loaded but were never written later on are removed from the linked lists */

  OTF_cleanup();

  OptRemoveUnreachableCode(stdout,0);

  inoptdata=1;

  // unreachable edge removal
  
  FOREACH_LIVE_FUN(ifun)
    {
      if (AFUN_IS_PSEUDO(ifun)) continue;
      
      FOREACH_FUN_BBL(ifun,ibbl)
	if (ABBL_TYPE(ibbl)==BT_BRANCH)
	  {
	    AEDG_UNMARK(ABBL_EDG_TRUE(ibbl));
	    AEDG_UNMARK(ABBL_EDG_FALSE(ibbl));
	  }

      if (0 && ifun>CONFIG_MIN && ifun<CONFIG_MAX)
	{
	  AfunIntraEdgSetCI(ifun);
	  AfunConstPropIntraEdg(ifun,-1);
	  AfunKillEdgesIntraEdg(ifun);
	}
      else
	{
	  AfunSetCI(ifun);
	  AfunConstProp(ifun);
	  AfunKillEdges(ifun);
	}

      FOREACH_FUN_BBL(ifun,ibbl)
	{
	  if (!ABBL_LIVE(ibbl) && ABBL_COUNT(ibbl)>0)
	    {
	      // QIUCK HACK: DON'T KNOW WHY IT WORKS !!!!!! BUT IT DOES WORK FOR ALL THE BENCHMARKS WE HAVE TESTED

	      //      printf("STRANGE supposedly dead block %d (%c) in fun %d with count %ld\n",ibbl,ABBL_TYPE(ibbl),ifun,ABBL_COUNT(ibbl));
	      continue;
	    }
	  if (0 && ifun>CONFIG_MIN && ifun<CONFIG_MAX)
	    {
	      AbblJoinIntraEdgValueSpecs(ibbl);
	    }
	  AbblSetValueSpec(ibbl);
	}
      
      ClearBblValueSpecsFun(ifun);
      ClearIntraEdgValueSpecsFun(ifun);
    }

  OTF_cleanup();

  FOREACH_LIVE_INS(iins)
    {
      AinsTrivialEvalNew(iins);
      RegisterForward(iins);
      
      switch( AINS_TYPE(iins) )
	{
	case IT_ICM:
	  StrRedICM(iins);
	  break;
	  
	case IT_FCM:
	  StrRedFCM(iins);
	  break;
	  
	case IT_IBR:
	  StrRedIBR(iins);
	  break;
	  
	case IT_FBR:
	  StrRedFBR(iins);
	  break;
	}
    }

  /* each calling context may be propagated into a procedure once more to do an evaluation */
#if LATE_CALL_JOIN
  ContextSensitiveElimination();
#endif      

  FOREACH_LIVE_INS(iins)    
    {
      AinsTrivialEvalNew(iins);
      RegisterForward(iins);
      switch( AINS_TYPE(iins) )
	{
	case IT_ICM:
	  StrRedICM(iins);
	  break;
	  
	case IT_FCM:
	  StrRedFCM(iins);
	  break;
	  
	case IT_IBR:
	  StrRedIBR(iins);
	  break;
	  
	case IT_FBR:
	  StrRedFBR(iins);
	  break;
	  
	case IT_JMP:
	  StrRedJMP(iins);
	  break;
	}
    }

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
  
  InfoStatus(fp,"\t[OptEval] {%d ms}\n",time);


  FOREACH_BBL(ibbl)
    ABBL_UNLIVE(ibbl);

  
  ComputeSavedChangedRegisters(stdout); // NECESSARY FOR PROPAGATION OF STACK-SAVED REGISTER CONTENTS.
  ComLiveness(stdout); // NECESSARY FOR PROPAGATION OF STACK-SAVED REGISTER CONTENTS.
  OptRemoveUnreachableCode(stdout,0);

}
/*==================================================================
  Mark instructions in a bbl that are not idempotent for the given
  input register values
  ===================================================================*/

static void AbblMarkValueSpec(INDEX ibbl)
{
  INDEX iins;
  VALUESPEC* input = GetBblValueSpec(ibbl);
  VALUESPEC tmp1;
  VALUESPEC* output =&tmp1;//(VALUESPEC *) malloc(sizeof(VALUESPEC));
  INT64 previous,next;
  REG reg;
  
  INT64 top;
  INT64 bot;

  CopyValueSpec(input,output);
  
  FOREACH_BBL_INS(ibbl,iins)
    {
      if (AINS_FLAGS(iins) & F_KILLED) continue;
      
      reg = AINS_REG_C(iins);
      top = output->top & (REG2BIT(reg));
      bot = output->bot & (REG2BIT(reg));
      previous = output->value[reg].i;
      AinsEvalValueSpecNew(iins,output);
      next = output->value[reg].i;
      top |= output->top & (REG2BIT(reg));
      bot |= output->bot & (REG2BIT(reg));
      
      if (reg != REG_ZERO && reg != REG_RA && reg != REG_FZERO && reg != REG_LIT && reg != REG_GP)
	if (!(bot|top))
	  if (next == previous)
	    continue;
      AINS_FLAGS(iins)|=F_MARKED;
    }
  //  free(output);
}
/*==================================================================
  For all instructions in a bbl, the definitive constants are set
  in the instructions
  ===================================================================*/

static void AbblSetValueSpec(INDEX ibbl)
{
  INDEX iins;
  VALUESPEC* input = GetBblValueSpec(ibbl);
  VALUESPEC tmp1;
  VALUESPEC tmp2;
  VALUESPEC* output = &tmp1;//(VALUESPEC *) malloc(sizeof(VALUESPEC));
  VALUESPEC* save =&tmp2;//(VALUESPEC *) malloc(sizeof(VALUESPEC));
  INT64 previous,next;
  REG reg;
#if ELIMINATION
  BOOL flag;
#endif
  
  INT64 top;
  INT64 bot;


  CopyValueSpec(input,output);

  FOREACH_BBL_INS(ibbl,iins)
    {
      if (AINS_FLAGS(iins) & F_KILLED) continue;
      
      reg = AINS_REG_A(iins);
      
      if ((AINS_TYPE(iins)!=IT_IST && AINS_TYPE(iins)!=IT_CST && AINS_TYPE(iins)!=IT_FST) || AINS_REG_B(iins)!=REG_SP || !ABBL_IS_FIRST(ibbl))
	if (reg != REG_LIT && reg != REG_ZERO && reg != REG_FZERO)
	  {
	    if (!( (output->top | output->bot) & REG2BIT(reg) ))
	      {
		AINS_A_EVAL(iins)=output->value[reg].i;
		AINS_FLAGS(iins) |= FI_EVAL_A;
	      }
	  }

      reg = AINS_REG_B(iins);
      
      if (reg != REG_LIT && reg != REG_ZERO && reg != REG_FZERO && reg != REG_GP)
	if (!( (output->top | output->bot) & REG2BIT(reg) ))
	  {
	    AINS_B_EVAL(iins)=output->value[reg].i;
	    AINS_FLAGS(iins) |= FI_EVAL_B;
	  }
      
      reg = AINS_REG_C(iins);
      
      CopyValueSpec(output,save);

      top = output->top & (REG2BIT(reg));
      bot = output->bot & (REG2BIT(reg));
      previous = output->value[reg].i;
      
      AinsEvalValueSpecNew(iins,output);
      
      next = output->value[reg].i;
      top |= output->top & (REG2BIT(reg));
      bot |= output->bot & (REG2BIT(reg));
      
      if (reg != REG_ZERO && reg != REG_FZERO)
	if (!( (output->top | output->bot) & REG2BIT(reg) ))
	  {
	    AINS_C_EVAL(iins)=output->value[reg].i;
	    AINS_FLAGS(iins) |= FI_EVAL_C;
	  }
#if ELIMINATION
      flag = 0;
      if (AINS_FLAGS(iins) & FI_EVAL_C) 
	{
	  flag = OptConstGenOneNew(stdout,ibbl,iins,save);
	}
      
      if (!flag && reg != REG_ZERO && reg != REG_RA && reg != REG_FZERO && reg != REG_LIT && reg != REG_GP)
	{
	  if (!(bot|top))
	    {
	      if (next == previous && !AddressIsCode(previous))
		{
		  AinsKill(iins);
		}
	    }
	}
#endif
    }
#if ELIMINATION
  OptConstGenTwoNew(stdout,ibbl);
#endif
  
  //  free(output);
  //  free(save);
  
}

/*==================================================================
  Propagate an outgoing register set to the successors of a bbl
  ===================================================================*/
static BOOL AbblProp2Succ(INDEX ibbl,VALUESPEC* bbl_output_spec)
{

  INDEX ifun = ABBL_FUN(ibbl);
  BOOL change = 0;
  REGSET regs;
  INDEX jbbl;
  INDEX last_iins;
  INDEX iedg;
  INT64 value = 0;
  double fvalue = 0.0;
  INDEX branch = INVALID;
  REG reg;
  VALUE Top;
  VALUE Bot;
  VALUESPEC tmp1;
  VALUESPEC tmp2;
  Top.i = TheLatticeTop;
  Bot.i = TheLatticeBot;

  switch(ABBL_TYPE(ibbl))
    {
    case BT_JUMP:
    case BT_EXIT:
    case BT_HALT:
      break;
    case BT_RET:
      jbbl=AFUN_BBL_LAST(ifun);
      if (GetBblValueSpec(jbbl))
	{
	  JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),AbblRegsLiveIn(jbbl));
	}
      ABBL_SET_LIVE(jbbl);
      break;
    case BT_PAL:
      jbbl=AEDG_TAIL(ABBL_EDG_LINK(ibbl));
      last_iins = ABBL_INS_LAST(ibbl);
      regs = /*REGSET_DIFF(REGS_ALL,REG2BIT(REG_V0));*/~AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)))); //NIET 100%  ZEKER !!!!
      if (GetBblValueSpec(jbbl))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  if (JoinValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),regs,AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    case BT_CALL:
      jbbl=AEDG_TAIL(ABBL_EDG_LINK(ibbl));
      last_iins = ABBL_INS_LAST(ibbl);

      if (AINS_CODE(last_iins)==I_jsr)
	{
	  REG reg = AINS_REG_B(last_iins);
	  if (!((bbl_output_spec->bot | bbl_output_spec->top) & REG2BIT(reg))
	      && LocateSection(bbl_output_spec->value[reg].i) == SysGetTextSection()
	      //	      && FindTargetBbl(bbl_output_spec->value[reg].i) != AEDG_TAIL(ABBL_EDG_CALL(ibbl))
	      && AFUN_IS_PSEUDO((ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)))))
	      && !(ABBL_FLAGS(ibbl)&FB_REEDGED)
	      )
	    {
	      INDEX target;
	      INT32 valuespec1,valuespec2;
	      INDEX repl_edg;
	      
	      target = FindTargetBbl(bbl_output_spec->value[reg].i);
	      //	      printf("changing block %d in fun %d to point to %d in fun %d\n",ibbl,ABBL_FUN(ibbl),target,ABBL_FUN(target));
	      valuespec1=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl));
	      valuespec2=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl)+1);
	      repl_edg = ABBL_EDG_CALL(ibbl);
	      AedgUnlinkFancy(repl_edg);
	      AEDG_TAIL(repl_edg)=target;
	      AedgLinkFancy(repl_edg);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl),valuespec1,Bot);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl)+1,valuespec2,Bot);
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl)+1);
	      AEDG_MARK(ABBL_EDG_CALL(ibbl));
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));
	      ABBL_FLAGS(ibbl) |= FB_REEDGED;
	    }
	  else if (!AFUN_IS_PSEUDO((ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)))))
		   && (((bbl_output_spec->bot | bbl_output_spec->top) & REG2BIT(reg))
		       || FindTargetBbl(bbl_output_spec->value[reg].i) != AEDG_TAIL(ABBL_EDG_CALL(ibbl))))
	    {
	      INDEX repl_edg;
	      INT32 valuespec1,valuespec2;
	      ABBL_FLAGS(ibbl) |= FB_REEDGED;
	      valuespec1=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl));
	      valuespec2=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl)+1);
	      repl_edg = ABBL_EDG_CALL(ibbl);
	      AedgUnlinkFancy(repl_edg);
	      AEDG_TAIL(repl_edg)=AFUN_BBL_FIRST(AFUN_PSEUDO_CALL);
	      AedgLinkFancy(repl_edg);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl),valuespec1,Top);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl)+1,valuespec2,Top);
	      //	      AEDG_MARK(ABBL_EDG_CALL(ibbl)+1);
	      //	      printf("changing block %d in fun %d to point to %d in fun %d\n",ibbl,ABBL_FUN(ibbl),AEDG_TAIL(ABBL_EDG_CALL(ibbl)),ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl)+1);
	      //	      AEDG_MARK(ABBL_EDG_CALL(ibbl));
	      //	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));
	    }
	  else if (((bbl_output_spec->bot | bbl_output_spec->top) & REG2BIT(reg)))
	    {
	      ABBL_FLAGS(ibbl) |= FB_REEDGED;
	    }

	}

      regs = ~AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));

      if (GetBblValueSpec(jbbl))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  if (JoinValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),regs,AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}

      break;

    case BT_SWITCH:
      FOREACH_BBL_SUCC(ibbl,iedg)
	{
	  if (AEDG_IS_KILLED(iedg)) continue;
	  jbbl=AEDG_TAIL(iedg);
	  if (GetBblValueSpec(jbbl))
	    {
	      if (!ABBL_LIVE(jbbl))
		{
		  ABBL_MARK(jbbl);
		}
	      ABBL_SET_LIVE(jbbl);
	      if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),AbblRegsLiveIn(jbbl)))
		{
		  change = 1;
		  ABBL_MARK(jbbl);
		}
	    }
	}
      break;
    case BT_NORM:  
      jbbl=AEDG_TAIL(ABBL_EDG_NORM(ibbl));
      if (GetBblValueSpec(jbbl))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    case BT_EXTRA:  
      jbbl=AEDG_TAIL(ABBL_EDG_EXTRA(ibbl));
      if (GetBblValueSpec(jbbl))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    case BT_BRANCH:
      last_iins = ABBL_INS_LAST(ibbl);
      value = bbl_output_spec->value[AINS_REG_A(last_iins)].i;

      if (CONDITIONAL_CONSTANT_PROPAGATION && !((bbl_output_spec->top|bbl_output_spec->bot) & REG2BIT(AINS_REG_A(last_iins))))
	{
	  if (AINS_TYPE(last_iins) == IT_IBR)
	    {
	      switch (AINS_CODE(last_iins))
		{
		case I_blbc:
		  if (!(value & 1L)) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_beq:
		  if (value == 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_blt:
		  if (value < 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_ble:
		  if (value <= 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_blbs:
		  if (value & 1L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_bne:
		  if (value != 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_bge:
		  if (value >= 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_bgt:
		  if (value > 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		}
	    }
	  else if (AINS_TYPE(last_iins) == IT_FBR)
	    {
	      fvalue =  bbl_output_spec->value[AINS_REG_A(last_iins)].f;
	      switch (AINS_CODE(last_iins))
		{
		case I_fbeq:
		  if (fvalue == (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fblt:
		  if (fvalue < (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fble:
		  if (fvalue <= (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fbne:
		  if (fvalue != (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fbge:
		  if (fvalue >= (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fbgt:
		  if (fvalue > (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		}
	    }
		  
		  
	  jbbl=AEDG_TAIL(branch);

	  AEDG_MARK(branch);
	  if (GetBblValueSpec(jbbl))
	    {
	      if (!ABBL_LIVE(jbbl))
		{
		  ABBL_MARK(jbbl);
		}	
	      ABBL_SET_LIVE(jbbl);
	      if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),AbblRegsLiveIn(jbbl)))
		{
		  change = 1;
		  ABBL_MARK(jbbl);
		}
	    }
	}
      else
	{
	  if( PREDICATE_CONSTANT_PROPAGATION && ( AINS_CODE(last_iins) == I_bne || AINS_CODE(last_iins) == I_beq ))
	    {
	      INDEX iins;
	      INDEX bbl_zero;
	      INDEX bbl_nonzero;
	      INDEX edg_zero;
	      INDEX edg_nonzero;
	      VALUESPEC* value_zero = &tmp1;//(VALUESPEC*) malloc (sizeof(VALUESPEC));
	      VALUESPEC* value_nonzero = &tmp2;//(VALUESPEC*) malloc (sizeof(VALUESPEC));
		      
	      REGSET spoiled = REGS_NONE;

	      if( AINS_CODE(last_iins) == I_bne )
		{
		  edg_zero = ABBL_EDG_FALSE(ibbl);
		  edg_nonzero = ABBL_EDG_TRUE(ibbl);
		}
	      else
		{
		  edg_nonzero = ABBL_EDG_FALSE(ibbl);
		  edg_zero = ABBL_EDG_TRUE(ibbl);
		}
		      
	      bbl_zero = AEDG_TAIL(edg_zero);
	      bbl_nonzero = AEDG_TAIL(edg_nonzero);
		      
	      CopyValueSpec(bbl_output_spec,value_zero);
	      CopyValueSpec(bbl_output_spec,value_nonzero);
		      
	      reg = AINS_REG_A(last_iins);
		      
	      spoiled = REGS_NONE;
		      
	      for(iins=AINS_PREV(last_iins);VALID(iins);iins = AINS_PREV(iins))
		{
		  spoiled |= AinsDef(iins);
		  if( AINS_REG_C(iins) == reg ) break;
		}
		      
	      value_zero->value[reg].i = 0L;
	      value_zero->top &= ~REG2BIT(reg);
	      value_zero->bot &= ~REG2BIT(reg);
		      


	      if( VALID(iins) && AINS_TYPE(iins) == IT_ICP && AINS_CODE(iins) != I_cmpbge )
		{
		  value_nonzero->value[reg].i = 1L;
		  value_nonzero->top &= ~REG2BIT(reg);
		  value_nonzero->bot &= ~REG2BIT(reg);
		}    

	      if (VALID(iins))
		{
		  REG rega = AINS_REG_A(iins);
		  REG regb = AINS_REG_B(iins);
			  
		  reg = AINS_REG_C(iins);
			  
		  if (AINS_FLAGS(iins) & FI_EVAL_B)
		    {
		      value = AINS_B_EVAL(iins);
		      reg = rega;
		    }
		  else if ((!REG2BIT(regb) & spoiled) && CONST(bbl_output_spec,regb))
		    {
		      value = bbl_output_spec->value[regb].i;
		      reg = rega;
		    }
		  else  if (AINS_FLAGS(iins) & FI_EVAL_A)
		    {
		      value = AINS_A_EVAL(iins);
		      reg = regb;
		    }
		  else if (!(REG2BIT(rega) & spoiled) && CONST(bbl_output_spec,rega))
		    {
		      value = bbl_output_spec->value[rega].i;
		      reg = regb;
		    }
		  
		  if (AINS_CODE(iins)==I_bis)
		    {
		      //		      if (AINS_REG_B(iins)==REG_LIT || AINS_FLAGS(iins)&FI_EVAL_B)
		      //			printf("opportunity to eliminate a branch in bbl %d\n",ibbl);
		      
		      if (!(REG2BIT(rega) & spoiled))
			{
			  value_zero->value[rega].i = 0L;
			  value_zero->top &= ~REG2BIT(rega);
			  value_zero->bot &= ~REG2BIT(rega);
			}
		      if (regb!=REG_LIT && !(REG2BIT(regb) & spoiled))
			{
			  value_zero->value[regb].i = 0L;
			  value_zero->top &= ~REG2BIT(regb);
			  value_zero->bot &= ~REG2BIT(regb);
			}
		    }
		  else if (!(REG2BIT(reg) & spoiled))
		    {
		      switch( AINS_CODE(iins) )
			{
			case I_cmpeq:
			  value_nonzero->value[reg].i = value;
			  value_nonzero->top &= ~REG2BIT(reg);
			  value_nonzero->bot &= ~REG2BIT(reg);
			  // MIGHT NEED EXTRA BITS SET AND RESET;
			  break;
			case I_addq:
			  value_zero->value[reg].i = -value;
			  value_zero->top &= ~REG2BIT(reg);
			  value_zero->bot &= ~REG2BIT(reg);
			  break;
			case I_subq:
			case I_xor:
			  value_zero->value[reg].i = value;
			  value_zero->top &= ~REG2BIT(reg);
			  value_zero->bot &= ~REG2BIT(reg);
			  break;
			case I_mull:
			case I_mulq:
			  if (value)
			    {
			      value_zero->value[reg].i = 0L;
			      value_zero->top &= ~REG2BIT(reg);
			      value_zero->bot &= ~REG2BIT(reg);
			    }
			  break;
			default:
			}
		    }
		}
		      
	      AEDG_MARK(edg_zero);
	      if (GetBblValueSpec(bbl_zero))
		{
		  if (!ABBL_LIVE(bbl_zero))
		    {
		      ABBL_MARK(bbl_zero);
		    }
		  ABBL_SET_LIVE(bbl_zero);
		  if (JoinAllValueSpec(value_zero,GetBblValueSpec(bbl_zero),AbblRegsLiveIn(bbl_zero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_zero);
		    }
		}

	      AEDG_MARK(edg_nonzero);
	      if (GetBblValueSpec(bbl_nonzero))
		{ 
		  if (!ABBL_LIVE(bbl_nonzero))
		    {
		      ABBL_MARK(bbl_nonzero);
		    }
		  ABBL_SET_LIVE(bbl_nonzero);
		  if (JoinAllValueSpec(value_nonzero,GetBblValueSpec(bbl_nonzero),AbblRegsLiveIn(bbl_nonzero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_nonzero);
		    }
		}
	      //	      free(value_zero);
	      //	      free(value_nonzero);
	    }
	  else 
	    {
	      INDEX edg_zero = ABBL_EDG_TRUE(ibbl);
	      INDEX edg_nonzero = ABBL_EDG_FALSE(ibbl);
	      INDEX bbl_zero = AEDG_TAIL(edg_zero);
	      INDEX bbl_nonzero = AEDG_TAIL(edg_nonzero);

	      AEDG_MARK(edg_zero);

	      if (GetBblValueSpec(bbl_zero))
		{
		  if (!ABBL_LIVE(bbl_zero))
		    {
		      ABBL_MARK(bbl_zero);
		    }
		  ABBL_SET_LIVE(bbl_zero);
		  if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(bbl_zero),AbblRegsLiveIn(bbl_zero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_zero);
		    }
		}
		      
	      AEDG_MARK(edg_nonzero);

	      if (GetBblValueSpec(bbl_nonzero))
		{
		  if (!ABBL_LIVE(bbl_nonzero))
		    {
		      ABBL_MARK(bbl_nonzero);
		    }
		  ABBL_SET_LIVE(bbl_nonzero);
		  if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(bbl_nonzero),AbblRegsLiveIn(bbl_nonzero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_nonzero);
		    }
		}
	    }
	}
	      
      break;
    case BT_UNCOND:  
      jbbl=AEDG_TAIL(ABBL_EDG_UNCOND(ibbl));

      if (GetBblValueSpec(jbbl))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  if (JoinAllValueSpec(bbl_output_spec,GetBblValueSpec(jbbl),AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    }

  return change;
}

static BOOL IntraEdgProp2Succ(INDEX ibbl,VALUESPEC* bbl_output_spec)
{

  INDEX ifun = ABBL_FUN(ibbl);
  BOOL change = 0;
  REGSET regs;
  INDEX jbbl;
  INDEX last_iins;
  INDEX iedg,jedg;
  INT64 value = 0;
  double fvalue = 0.0;
  INDEX branch = INVALID;
  REG reg;
  VALUE Top;
  VALUE Bot;
  VALUESPEC tmp1;
  VALUESPEC tmp2;


  Top.i = TheLatticeTop;
  Bot.i = TheLatticeBot;

  switch(ABBL_TYPE(ibbl))
    {
    case BT_JUMP:
    case BT_EXIT:
    case BT_HALT:
      break;
    case BT_RET:
      jedg=ABBL_SUCC(ibbl);
      jbbl=AEDG_TAIL(jedg);
      if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	{
	  JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),AbblRegsLiveIn(jbbl));
	}
      ABBL_SET_LIVE(jbbl);
      AEDG_SET_LIVE(jedg);
      break;
    case BT_PAL:
      jbbl=AEDG_TAIL(ABBL_EDG_LINK(ibbl));
      jedg=ABBL_EDG_LINK(ibbl);
      last_iins = ABBL_INS_LAST(ibbl);
      regs = /*REGSET_DIFF(REGS_ALL,REG2BIT(REG_V0));*/~AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)))); //NIET 100%  ZEKER !!!!
      if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  AEDG_SET_LIVE(jedg);
	  if (JoinValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),regs,AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    case BT_CALL:
      jbbl=AEDG_TAIL(ABBL_EDG_LINK(ibbl));
      jedg=ABBL_EDG_LINK(ibbl);
      last_iins = ABBL_INS_LAST(ibbl);
      
      if (AINS_CODE(last_iins)==I_jsr)
	{
	  REG reg = AINS_REG_B(last_iins);
	  if (!((bbl_output_spec->bot | bbl_output_spec->top) & REG2BIT(reg))
	      && LocateSection(bbl_output_spec->value[reg].i) == SysGetTextSection()
	      //	      && FindTargetBbl(bbl_output_spec->value[reg].i) != AEDG_TAIL(ABBL_EDG_CALL(ibbl))
	      && AFUN_IS_PSEUDO((ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)))))
	      && !(ABBL_FLAGS(ibbl)&FB_REEDGED)
	      )
	    {
	      INDEX target;
	      INT32 valuespec1,valuespec2;
	      INDEX repl_edg;
	      
	      target = FindTargetBbl(bbl_output_spec->value[reg].i);
	      //	      printf("changing block %d in fun %d to point to %d in fun %d\n",ibbl,ABBL_FUN(ibbl),target,ABBL_FUN(target));
	      valuespec1=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl));
	      valuespec2=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl)+1);
	      repl_edg = ABBL_EDG_CALL(ibbl);
	      AedgUnlinkFancy(repl_edg);
	      AEDG_TAIL(repl_edg)=target;
	      AedgLinkFancy(repl_edg);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl),valuespec1,Bot);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl)+1,valuespec2,Bot);
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl)+1);
	      AEDG_MARK(ABBL_EDG_CALL(ibbl));
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));
	      ABBL_FLAGS(ibbl) |= FB_REEDGED;
	    }
	  else if (!AFUN_IS_PSEUDO((ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)))))
		   && (((bbl_output_spec->bot | bbl_output_spec->top) & REG2BIT(reg))
		       || FindTargetBbl(bbl_output_spec->value[reg].i) != AEDG_TAIL(ABBL_EDG_CALL(ibbl))))
	    {
	      INDEX repl_edg;
	      INT32 valuespec1,valuespec2;
	      ABBL_FLAGS(ibbl) |= FB_REEDGED;
	      valuespec1=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl));
	      valuespec2=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl)+1);
	      repl_edg = ABBL_EDG_CALL(ibbl);
	      AedgUnlinkFancy(repl_edg);
	      AEDG_TAIL(repl_edg)=AFUN_BBL_FIRST(AFUN_PSEUDO_CALL);
	      AedgLinkFancy(repl_edg);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl),valuespec1,Top);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl)+1,valuespec2,Top);
	      //	      AEDG_MARK(ABBL_EDG_CALL(ibbl)+1);
	      //	      printf("changing block %d in fun %d to point to %d in fun %d\n",ibbl,ABBL_FUN(ibbl),AEDG_TAIL(ABBL_EDG_CALL(ibbl)),ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl)+1);
	      //	      AEDG_MARK(ABBL_EDG_CALL(ibbl));
	      //	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));
	    }
	  else if (((bbl_output_spec->bot | bbl_output_spec->top) & REG2BIT(reg)))
	    {
	      ABBL_FLAGS(ibbl) |= FB_REEDGED;
	    }
	  
	}

      regs = ~AFUN_REGS_CHANGED(ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl))));

      if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  AEDG_SET_LIVE(jedg);

	  if (JoinValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),regs,AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      
      break;
      
    case BT_SWITCH:
      FOREACH_BBL_SUCC(ibbl,jedg)
	{
	  if (AEDG_IS_KILLED(jedg)) continue;
	  jbbl=AEDG_TAIL(jedg);
	  if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	    {
	      if (!ABBL_LIVE(jbbl))
		{
		  ABBL_MARK(jbbl);
		}
	      ABBL_SET_LIVE(jbbl);
	      AEDG_SET_LIVE(jedg);
	      if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),AbblRegsLiveIn(jbbl)))
		{
		  change = 1;
		  ABBL_MARK(jbbl);
		}
	    }
	}
      break;
    case BT_NORM:  
      jedg=ABBL_EDG_NORM(ibbl);
      jbbl=AEDG_TAIL(ABBL_EDG_NORM(ibbl));
      if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  AEDG_SET_LIVE(jedg);
	  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    case BT_EXTRA:  
      jedg=ABBL_EDG_EXTRA(ibbl);
      jbbl=AEDG_TAIL(ABBL_EDG_EXTRA(ibbl));
      if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  AEDG_SET_LIVE(jedg);
	  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    case BT_BRANCH:
      last_iins = ABBL_INS_LAST(ibbl);
      value = bbl_output_spec->value[AINS_REG_A(last_iins)].i;

      if (CONDITIONAL_CONSTANT_PROPAGATION && !((bbl_output_spec->top|bbl_output_spec->bot) & REG2BIT(AINS_REG_A(last_iins))))
	{
	  if (AINS_TYPE(last_iins) == IT_IBR)
	    {
	      switch (AINS_CODE(last_iins))
		{
		case I_blbc:
		  if (!(value & 1L)) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_beq:
		  if (value == 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_blt:
		  if (value < 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_ble:
		  if (value <= 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_blbs:
		  if (value & 1L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_bne:
		  if (value != 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_bge:
		  if (value >= 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_bgt:
		  if (value > 0L) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		}
	    }
	  else if (AINS_TYPE(last_iins) == IT_FBR)
	    {
	      fvalue =  bbl_output_spec->value[AINS_REG_A(last_iins)].f;
	      switch (AINS_CODE(last_iins))
		{
		case I_fbeq:
		  if (fvalue == (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fblt:
		  if (fvalue < (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fble:
		  if (fvalue <= (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fbne:
		  if (fvalue != (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fbge:
		  if (fvalue >= (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		case I_fbgt:
		  if (fvalue > (double) 0.0) branch = ABBL_EDG_TRUE(ibbl);
		  else branch = ABBL_EDG_FALSE(ibbl);
		  break;
		}
	    }
		  
		  
	  jbbl=AEDG_TAIL(branch);

	  AEDG_MARK(branch);
	  if (GetEdgValueSpec(branch) && !AEDG_IS_INTER(branch))
	    {
	      if (!ABBL_LIVE(jbbl))
		{
		  ABBL_MARK(jbbl);
		}	
	      ABBL_SET_LIVE(jbbl);
	      AEDG_SET_LIVE(branch);
	      if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(branch),AbblRegsLiveIn(jbbl)))
		{
		  change = 1;
		  ABBL_MARK(jbbl);
		}
	    }
	}
      else
	{
	  if( PREDICATE_CONSTANT_PROPAGATION && ( AINS_CODE(last_iins) == I_bne || AINS_CODE(last_iins) == I_beq ))
	    {
	      INDEX iins;
	      INDEX bbl_zero;
	      INDEX bbl_nonzero;
	      INDEX edg_zero;
	      INDEX edg_nonzero;
	      VALUESPEC* value_zero = &tmp1;//(VALUESPEC*) malloc (sizeof(VALUESPEC));
	      VALUESPEC* value_nonzero = &tmp2;//(VALUESPEC*) malloc (sizeof(VALUESPEC));
		      
	      REGSET spoiled = REGS_NONE;

	      if( AINS_CODE(last_iins) == I_bne )
		{
		  edg_zero = ABBL_EDG_FALSE(ibbl);
		  edg_nonzero = ABBL_EDG_TRUE(ibbl);
		}
	      else
		{
		  edg_nonzero = ABBL_EDG_FALSE(ibbl);
		  edg_zero = ABBL_EDG_TRUE(ibbl);
		}
		      
	      bbl_zero = AEDG_TAIL(edg_zero);
	      bbl_nonzero = AEDG_TAIL(edg_nonzero);
		      
	      CopyValueSpec(bbl_output_spec,value_zero);
	      CopyValueSpec(bbl_output_spec,value_nonzero);
		      
	      reg = AINS_REG_A(last_iins);
		      
	      spoiled = REGS_NONE;
		      
	      for(iins=AINS_PREV(last_iins);VALID(iins);iins = AINS_PREV(iins))
		{
		  spoiled |= AinsDef(iins);
		  if( AINS_REG_C(iins) == reg ) break;
		}
		      
	      value_zero->value[reg].i = 0L;
	      value_zero->top &= ~REG2BIT(reg);
	      value_zero->bot &= ~REG2BIT(reg);
		      


	      if( VALID(iins) && AINS_TYPE(iins) == IT_ICP && AINS_CODE(iins) != I_cmpbge )
		{
		  value_nonzero->value[reg].i = 1L;
		  value_nonzero->top &= ~REG2BIT(reg);
		  value_nonzero->bot &= ~REG2BIT(reg);
		}    

	      if (VALID(iins))
		{
		  REG rega = AINS_REG_A(iins);
		  REG regb = AINS_REG_B(iins);
			  
		  reg = AINS_REG_C(iins);
			  
		  if (AINS_FLAGS(iins) & FI_EVAL_B)
		    {
		      value = AINS_B_EVAL(iins);
		      reg = rega;
		    }
		  else if ((!REG2BIT(regb) & spoiled) && CONST(bbl_output_spec,regb))
		    {
		      value = bbl_output_spec->value[regb].i;
		      reg = rega;
		    }
		  else  if (AINS_FLAGS(iins) & FI_EVAL_A)
		    {
		      value = AINS_A_EVAL(iins);
		      reg = regb;
		    }
		  else if (!(REG2BIT(rega) & spoiled) && CONST(bbl_output_spec,rega))
		    {
		      value = bbl_output_spec->value[rega].i;
		      reg = regb;
		    }
		  
		  if (AINS_CODE(iins)==I_bis)
		    {
		      //		      if (AINS_REG_B(iins)==REG_LIT || AINS_FLAGS(iins)&FI_EVAL_B)
		      //			printf("opportunity to eliminate a branch in bbl %d\n",ibbl);
		      
		      if (!(REG2BIT(rega) & spoiled))
			{
			  value_zero->value[rega].i = 0L;
			  value_zero->top &= ~REG2BIT(rega);
			  value_zero->bot &= ~REG2BIT(rega);
			}
		      if (regb!=REG_LIT && !(REG2BIT(regb) & spoiled))
			{
			  value_zero->value[regb].i = 0L;
			  value_zero->top &= ~REG2BIT(regb);
			  value_zero->bot &= ~REG2BIT(regb);
			}
		    }
		  else if (!(REG2BIT(reg) & spoiled))
		    {
		      switch( AINS_CODE(iins) )
			{
			case I_cmpeq:
			  value_nonzero->value[reg].i = value;
			  value_nonzero->top &= ~REG2BIT(reg);
			  value_nonzero->bot &= ~REG2BIT(reg);
			  // MIGHT NEED EXTRA BITS SET AND RESET;
			  break;
			case I_addq:
			  value_zero->value[reg].i = -value;
			  value_zero->top &= ~REG2BIT(reg);
			  value_zero->bot &= ~REG2BIT(reg);
			  break;
			case I_subq:
			case I_xor:
			  value_zero->value[reg].i = value;
			  value_zero->top &= ~REG2BIT(reg);
			  value_zero->bot &= ~REG2BIT(reg);
			  break;
			case I_mull:
			case I_mulq:
			  if (value)
			    {
			      value_zero->value[reg].i = 0L;
			      value_zero->top &= ~REG2BIT(reg);
			      value_zero->bot &= ~REG2BIT(reg);
			    }
			  break;
			default:
			}
		    }
		}
		      
	      AEDG_MARK(edg_zero);
	      if (GetEdgValueSpec(edg_zero) && !AEDG_IS_INTER(edg_zero))
		{
		  if (!ABBL_LIVE(bbl_zero))
		    {
		      ABBL_MARK(bbl_zero);
		    }
		  ABBL_SET_LIVE(bbl_zero);
		  AEDG_SET_LIVE(edg_zero);
		  if (JoinAllValueSpec(value_zero,GetEdgValueSpec(edg_zero),AbblRegsLiveIn(bbl_zero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_zero);
		    }
		}

	      AEDG_MARK(edg_nonzero);
	      if (GetEdgValueSpec(edg_nonzero) && !AEDG_IS_INTER(edg_nonzero))
		{ 
		  if (!ABBL_LIVE(bbl_nonzero))
		    {
		      ABBL_MARK(bbl_nonzero);
		    }
		  ABBL_SET_LIVE(bbl_nonzero);
		  AEDG_SET_LIVE(edg_nonzero);
		  if (JoinAllValueSpec(value_nonzero,GetEdgValueSpec(edg_nonzero),AbblRegsLiveIn(bbl_nonzero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_nonzero);
		    }
		}
	      //	      free(value_zero);
	      //	      free(value_nonzero);
	    }
	  else 
	    {
	      INDEX edg_zero = ABBL_EDG_TRUE(ibbl);
	      INDEX edg_nonzero = ABBL_EDG_FALSE(ibbl);
	      INDEX bbl_zero = AEDG_TAIL(edg_zero);
	      INDEX bbl_nonzero = AEDG_TAIL(edg_nonzero);

	      AEDG_MARK(edg_zero);

	      if (GetEdgValueSpec(edg_zero) && !AEDG_IS_INTER(edg_zero))
		{
		  if (!ABBL_LIVE(bbl_zero))
		    {
		      ABBL_MARK(bbl_zero);
		    }
		  ABBL_SET_LIVE(bbl_zero);
		  AEDG_SET_LIVE(edg_zero);
		  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(edg_zero),AbblRegsLiveIn(bbl_zero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_zero);
		    }
		}
		      
	      AEDG_MARK(edg_nonzero);

	      if (GetEdgValueSpec(edg_nonzero) && !AEDG_IS_INTER(edg_nonzero))
		{
		  if (!ABBL_LIVE(bbl_nonzero))
		    {
		      ABBL_MARK(bbl_nonzero);
		    }
		  ABBL_SET_LIVE(bbl_nonzero);
		  AEDG_SET_LIVE(edg_nonzero);
		  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(edg_nonzero),AbblRegsLiveIn(bbl_nonzero)))
		    {
		      change = 1;
		      ABBL_MARK(bbl_nonzero);
		    }
		}
	    }
	}
	      
      break;
    case BT_UNCOND:  
      jedg=ABBL_EDG_UNCOND(ibbl);
      jbbl=AEDG_TAIL(ABBL_EDG_UNCOND(ibbl));

      if (GetEdgValueSpec(jedg) && !AEDG_IS_INTER(jedg))
	{
	  if (!ABBL_LIVE(jbbl))
	    {
	      ABBL_MARK(jbbl);
	    }
	  ABBL_SET_LIVE(jbbl);
	  AEDG_SET_LIVE(jedg);
	  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(jedg),AbblRegsLiveIn(jbbl)))
	    {
	      change = 1;
	      ABBL_MARK(jbbl);
	    }
	}
      break;
    }

  return change;
}

/*=============================================================
  Do a fixpoint over one procedure, needs AfunSetCI or AfunSetCS first
  ===============================================================*/
static void AfunConstProp(INDEX ifun)
{
  BOOL change; 
  INDEX ibbl;
  VALUESPEC tmp1;
  VALUESPEC* bbl_output_spec = &tmp1;//(VALUESPEC *) malloc (sizeof(VALUESPEC));
  change = 1;

  while (change)
    {
      change = 0;
      FOREACH_FUN_BBL(ifun,ibbl)
	{
	  if (ABBL_IS_KILLED(ibbl)) continue;
	  if (!ABBL_IS_MARKED(ibbl)) continue;
	  
	  ABBL_SET_LIVE(ibbl);
	  ABBL_UNMARK(ibbl);
	  
	  AbblEvalValueSpec(ibbl,bbl_output_spec);
	  change |= AbblProp2Succ(ibbl,bbl_output_spec);
	}
    }

  //  free(bbl_output_spec);
}

static void AfunConstPropIntraEdg(INDEX ifun, INDEX call)
{
  BOOL change; 
  INDEX ibbl;
  INDEX iedg;
  VALUESPEC tmp1;
  VALUESPEC* bbl_output_spec = &tmp1;//(VALUESPEC *) malloc (sizeof(VALUESPEC));
  change = 1;

  while (change)
    {
      change = 0;
      FOREACH_FUN_BBL(ifun,ibbl)
	{
	  if (ABBL_IS_KILLED(ibbl)) continue;
	  if (!ABBL_IS_MARKED(ibbl)) continue;

	  ABBL_SET_LIVE(ibbl);
	  ABBL_UNMARK(ibbl);
	  if (!VALID(call) || ibbl!=AFUN_BBL_FIRST(ifun)) // CI or CS on non-first block
	    {
	      FOREACH_BBL_PRED(ibbl,iedg)
		{
		  if (!AEDG_LIVE(iedg)) continue;
		  if (AEDG_TYPE(iedg)==ET_FUN_RETURN) continue;
		  IntraEdgEvalValueSpec(iedg,bbl_output_spec);
		  change |= IntraEdgProp2Succ(ibbl,bbl_output_spec);
		}
	    }
	  else //CS on first block
	    {
	      FOREACH_BBL_PRED(ibbl,iedg)
		{
		  if (AEDG_IS_INTER(iedg) && iedg!=call) continue;
		  if (!AEDG_LIVE(iedg)) continue;
		  if (AEDG_TYPE(iedg)==ET_FUN_RETURN) continue;
		  IntraEdgEvalValueSpec(iedg,bbl_output_spec);
		  change |= IntraEdgProp2Succ(ibbl,bbl_output_spec);
		}
	    }
	}
    }

  //  free(bbl_output_spec);
}

/*=============================================================
  Propagate the outgoing values after a proc's fixpoint for
  call-site UNspecific CP
  ===============================================================*/
#if PROPAGATION_METHOD
static BOOL AfunChangeCI(INDEX ifun)
{
  VALUESPEC tmp1;
  VALUESPEC* bbl_output_spec = &tmp1;//(VALUESPEC *) malloc (sizeof(VALUESPEC));
  INDEX ibbl;
  BOOL change=0; 
  INDEX succ;
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (!ABBL_LIVE(ibbl)) continue;
      
      AbblEvalValueSpec(ibbl,bbl_output_spec);
      
      FOREACH_BBL_SUCC(ibbl,succ)
	{
	  if (!AEDG_IS_INTER(succ)) continue;
	  if (!AEDG_LIVE(succ))
	    { 
	      AEDG_MARK(succ);
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
	      change = 1;
	    }
	  AEDG_SET_LIVE(succ);
	  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(succ),AbblRegsLiveIn(AEDG_TAIL(succ))))
	    {
	      AEDG_MARK(succ);
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
	      change = 1;
	    }
	}
    }
  //  free(bbl_output_spec);
  return change;
}

static BOOL AfunIntraEdgChangeCI(INDEX ifun)
{
  VALUESPEC tmp1;
  VALUESPEC* bbl_output_spec = &tmp1;//(VALUESPEC *) malloc (sizeof(VALUESPEC));
  INDEX ibbl;
  INDEX iedg;
  BOOL change=0; 
  INDEX succ;

  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (!ABBL_LIVE(ibbl)) continue;
      
      FOREACH_BBL_PRED(ibbl,iedg)
	{
	  if (!AEDG_LIVE(iedg)) continue;
	  if (AEDG_TYPE(iedg)==ET_FUN_RETURN) continue;
	  IntraEdgEvalValueSpec(iedg,bbl_output_spec);

	  FOREACH_BBL_SUCC(ibbl,succ)
	    {
	      if (!AEDG_IS_INTER(succ)) continue;
	      if (!AEDG_LIVE(succ))
		{ 
		  AEDG_MARK(succ);
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
		  change = 1;
		}
	      AEDG_SET_LIVE(succ);

	      if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(succ),AbblRegsLiveIn(AEDG_TAIL(succ))))
		{
		  AEDG_MARK(succ);
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
		  change = 1;
		}
	    }
	}
    }
  //  free(bbl_output_spec);
  return change;
}
#else
/*=============================================================
  same for call-site specific
  ===============================================================*/
static BOOL AfunChangeCS(INDEX ifun, INDEX call)
{
  VALUESPEC tmp1;
  VALUESPEC* bbl_output_spec = &tmp1;//(VALUESPEC *) malloc (sizeof(VALUESPEC));
  INDEX ibbl;
  BOOL change=0; 
  INDEX succ;
  INDEX pred;
  VALUE Top;
  VALUE Bot;
  Top.i = TheLatticeTop;
  Bot.i = TheLatticeBot;

  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (!ABBL_LIVE(ibbl)) continue;
      
      AbblEvalValueSpec(ibbl,bbl_output_spec);
      
      /*      if (ABBL_TYPE(ibbl)==BT_CALL)
	if (AINS_CODE(ABBL_INS_LAST(ibbl))==I_jsr)
	  if (!AFUN_IS_PSEUDO(AEDG_TAIL(ABBL_EDG_CALL(ibbl))))
	    {
	      INT32 valuespec1,valuespec2;
	      INDEX repl_edg;
	      valuespec1=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl));
	      valuespec2=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl)+1);
	      //	      AedgKillFancy(ABBL_EDG_CALL(ibbl));
	      //	      AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_FIRST(AFUN_PSEUDO_CALL),ET_FUN_CALL));
	      repl_edg = ABBL_EDG_CALL(ibbl);
	      AedgUnlinkFancy(repl_edg);
	      AEDG_TAIL(repl_edg)=AFUN_BBL_FIRST(AFUN_PSEUDO_CALL);
	      AedgLinkFancy(repl_edg);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl),valuespec1,Top);
	      ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl)+1,valuespec2,Top);
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl));
	      AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl)+1);
	    }
      */

      FOREACH_BBL_SUCC(ibbl,succ)
	{
	  if (!AEDG_IS_INTER(succ)) continue;
	  if (AEDG_TYPE(succ) == ET_FUN_RETURN && AEDG_TAIL(succ) != AEDG_TAIL(ABBL_EDG_LINK(AEDG_HEAD(call)))) continue;
	  if (!AEDG_LIVE(succ))
	    { 
	      AEDG_MARK(succ);
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
	      change = 1;
	    }
	  AEDG_SET_LIVE(succ);

	  if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(succ),AbblRegsLiveIn(AEDG_TAIL(succ))))
	    {
	      AEDG_MARK(succ);
	      AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
	      if (AEDG_TYPE(succ) != ET_FUN_CALL)
		{
		  FOREACH_BBL_PRED(AFUN_BBL_FIRST(ABBL_FUN(AEDG_TAIL(succ))),pred)
		    if (AEDG_IS_INTER(pred))
		      AEDG_MARK(pred);
		}
	      change = 1;
	    }
	}
    }
  //  free(bbl_output_spec);
  return change;
}

static BOOL AfunIntraEdgChangeCS(INDEX ifun, INDEX call)
{
  VALUESPEC tmp1;
  VALUESPEC* bbl_output_spec = &tmp1;//(VALUESPEC *) malloc (sizeof(VALUESPEC));
  INDEX ibbl;
  BOOL change=0; 
  INDEX succ;
  INDEX pred;
  VALUE Top;
  VALUE Bot;
  INDEX iedg;

  Top.i = TheLatticeTop;
  Bot.i = TheLatticeBot;

  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (!ABBL_LIVE(ibbl)) continue;
      
      FOREACH_BBL_PRED(ibbl,iedg)
	{
	  if (!AEDG_LIVE(iedg)) continue;
	  if (AEDG_TYPE(iedg)==ET_FUN_RETURN) continue;
	  IntraEdgEvalValueSpec(iedg,bbl_output_spec);
      
	  /*      if (ABBL_TYPE(ibbl)==BT_CALL)
		  if (AINS_CODE(ABBL_INS_LAST(ibbl))==I_jsr)
		  if (!AFUN_IS_PSEUDO(AEDG_TAIL(ABBL_EDG_CALL(ibbl))))
		  {
		  INT32 valuespec1,valuespec2;
		  INDEX repl_edg;
		  valuespec1=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl));
		  valuespec2=AEDG_VALUESPEC(ABBL_EDG_CALL(ibbl)+1);
		  //	      AedgKillFancy(ABBL_EDG_CALL(ibbl));
		  //	      AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_FIRST(AFUN_PSEUDO_CALL),ET_FUN_CALL));
		  repl_edg = ABBL_EDG_CALL(ibbl);
		  AedgUnlinkFancy(repl_edg);
		  AEDG_TAIL(repl_edg)=AFUN_BBL_FIRST(AFUN_PSEUDO_CALL);
		  AedgLinkFancy(repl_edg);
		  ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl),valuespec1,Top);
		  ReuseEdgValueSpec(ABBL_EDG_CALL(ibbl)+1,valuespec2,Top);
		  AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl));
		  AEDG_SET_LIVE(ABBL_EDG_CALL(ibbl)+1);
		  }
	  */
	  
	  FOREACH_BBL_SUCC(ibbl,succ)
	    {
	      if (!AEDG_IS_INTER(succ)) continue;
	      if (ABBL_TYPE(ibbl)==BT_BRANCH && !((bbl_output_spec->top | bbl_output_spec->bot) & REG2BIT(AINS_REG_A(ABBL_INS_LAST(ibbl)))))
		if (!AEDG_LIVE(succ))
		  {
		    continue;
		  }
	      if (AEDG_TYPE(succ) == ET_FUN_RETURN && AEDG_TAIL(succ) != AEDG_TAIL(ABBL_EDG_LINK(AEDG_HEAD(call)))) continue;
	      if (!AEDG_LIVE(succ))
		{ 
		  AEDG_MARK(succ);
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
		  change = 1;
		}

	      AEDG_SET_LIVE(succ);
	      
	      if (JoinAllValueSpec(bbl_output_spec,GetEdgValueSpec(succ),AbblRegsLiveIn(AEDG_TAIL(succ))))
		{
		  AEDG_MARK(succ);
		  AFUN_MARK(ABBL_FUN(AEDG_TAIL(succ)));
		  if (AEDG_TYPE(succ) != ET_FUN_CALL)
		    {
		      FOREACH_BBL_PRED(AFUN_BBL_FIRST(ABBL_FUN(AEDG_TAIL(succ))),pred)
			if (AEDG_IS_INTER(pred))
			  AEDG_MARK(pred);
		    }
		  change = 1;
		}
	    }
	}
    }
  //  free(bbl_output_spec);





  return change;
}

#endif
/*=============================================================
  Initialize the values for a procedures fixpoint
  ===============================================================*/
static void AfunSetCI(INDEX ifun)
{
  INDEX ibbl;
  INDEX pred;
  VALUE Bot;
  VALUE Top;
  VALUE Gp;
  VALUE Zero;
  VALUE FZero;
  BOOL flag;
  REG reg;

  INT64 LatticeBot = TheLatticeBot;
  INT64 LatticeTop = TheLatticeTop;
  Bot.i = LatticeBot;
  Top.i = LatticeTop;
  Gp.i = SysGetGlobalPointer();
  Zero.i = 0;
  FZero.i = 0.0;
  Gp.i = SysGetGlobalPointer();

  ClearBblValueSpecsFun(ifun);
      
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (ABBL_IS_KILLED(ibbl)) continue;
      if (BBL_CONSTANT_PROPAGATION)
	{
	  NewBblValueSpec(ibbl,Bot);
#if !PROP_DEAD_VALUES
	  REGSET_FOREACH_REG(~AbblRegsLiveIn(ibbl),reg)
	    SetOneBblValueSpec(ibbl,Top,reg);
	  SetOneBblValueSpec(ibbl,Bot,REG_GP);
	  SetOneBblValueSpec(ibbl,Bot,REG_ZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_FZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_RA);
	  SetOneBblValueSpec(ibbl,Bot,REG_SP);
	  SetOneBblValueSpec(ibbl,Bot,REG_PV);
	  SetOneBblValueSpec(ibbl,Bot,REG_AT);
#endif
	}
      else 
	{
	  NewBblValueSpec(ibbl,Top);
	  SetOneBblValueSpec(ibbl,Bot,REG_GP);
	  SetOneBblValueSpec(ibbl,Bot,REG_ZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_FZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_RA);
	  SetOneBblValueSpec(ibbl,Bot,REG_SP);
	  SetOneBblValueSpec(ibbl,Bot,REG_PV);
	  SetOneBblValueSpec(ibbl,Bot,REG_AT);
	}
      ABBL_UNMARK(ibbl);
      ABBL_UNLIVE(ibbl);
      flag = 0;
      FOREACH_BBL_PRED(ibbl,pred)
	{
	  if (AEDG_IS_INTER(pred) && AEDG_LIVE(pred))
	    {
	      JoinValueSpec(GetEdgValueSpec(pred),GetBblValueSpec(ibbl),
			    AEDG_TYPE(pred)==ET_FUN_RETURN?
			    (
			     /*AFUN_PSEUDO_PAL == ABBL_FUN(AEDG_HEAD(pred))?REG2BIT(REG_V0):*/AFUN_REGS_CHANGED(ABBL_FUN(AEDG_HEAD(pred))) // NIET 100% ZEKER
			     )
			    :REGS_ALL
			    ,AbblRegsLiveIn(ibbl)
			    );
	      ABBL_SET_LIVE(ibbl);
	      ABBL_MARK(ibbl);
	    }
	  if (AEDG_TYPE(pred)==ET_FUN_LINK && !AEDG_IS_INTER(pred))
	    flag = 1;
	}
      if (flag)
	{
	  ABBL_UNMARK(ibbl);
	  ABBL_UNLIVE(ibbl);
	}
    }
}

static void AfunIntraEdgSetCI(INDEX ifun)
{
  INDEX ibbl;
  INDEX pred;
  VALUE Bot;
  VALUE Top;
  VALUE Gp;
  VALUE Zero;
  VALUE FZero;
  BOOL flag;
  REG reg;
  INDEX iedg;
  INT64 LatticeBot = TheLatticeBot;
  INT64 LatticeTop = TheLatticeTop;
  Bot.i = LatticeBot;
  Top.i = LatticeTop;
  Gp.i = SysGetGlobalPointer();
  Zero.i = 0;
  FZero.i = 0.0;
  Gp.i = SysGetGlobalPointer();
  ClearIntraEdgValueSpecsFun(ifun);
      
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      ABBL_UNMARK(ibbl);
      ABBL_UNLIVE(ibbl);
      FOREACH_BBL_SUCC(ibbl,iedg)
	{
	  if (!AEDG_IS_INTER(iedg))
	    {
	      AEDG_UNLIVE(iedg);
	      if (BBL_CONSTANT_PROPAGATION)
		{
		  NewIntraEdgValueSpec(iedg,Bot);
#if !PROP_DEAD_VALUES
		  REGSET_FOREACH_REG(~AbblRegsLiveIn(AEDG_TAIL(iedg)),reg)
		    SetOneEdgValueSpec(iedg,Top,reg);
		  SetOneEdgValueSpec(iedg,Bot,REG_GP);
		  SetOneEdgValueSpec(iedg,Bot,REG_ZERO);
		  SetOneEdgValueSpec(iedg,Bot,REG_FZERO);
		  SetOneEdgValueSpec(iedg,Bot,REG_RA);
		  SetOneEdgValueSpec(iedg,Bot,REG_SP);
		  SetOneEdgValueSpec(iedg,Bot,REG_PV);
		  SetOneEdgValueSpec(iedg,Bot,REG_AT);
#endif
		}
	      else 
		{
		  NewIntraEdgValueSpec(iedg,Top);
		  SetOneEdgValueSpec(iedg,Bot,REG_GP);
		  SetOneEdgValueSpec(iedg,Bot,REG_ZERO);
		  SetOneEdgValueSpec(iedg,Bot,REG_FZERO);
		  SetOneEdgValueSpec(iedg,Bot,REG_RA);
		  SetOneEdgValueSpec(iedg,Bot,REG_SP);
		  SetOneEdgValueSpec(iedg,Bot,REG_PV);
		  SetOneEdgValueSpec(iedg,Bot,REG_AT);
		}
	    }
	  flag = 0;
	  FOREACH_BBL_PRED(ibbl,pred)
	    {
	      if (AEDG_IS_INTER(pred) && AEDG_LIVE(pred))
		{
		  ABBL_SET_LIVE(ibbl);
		  ABBL_MARK(ibbl);
		}
	      if (AEDG_TYPE(pred)==ET_FUN_LINK && !AEDG_IS_INTER(pred))
		flag = 1;
	    }
	  if (flag)
	    {
	      ABBL_UNMARK(ibbl);
	      ABBL_UNLIVE(ibbl);
	    }	
	}
    }
  
  ABBL_MARK(AFUN_BBL_FIRST(ibbl));
  ABBL_SET_LIVE(AFUN_BBL_FIRST(ibbl));
}

/*=============================================================
  same for context sensitive
  ===============================================================*/

static void AfunSetCS(INDEX ifun, INDEX call)
{
  INDEX ibbl;
  INDEX pred;
  VALUE Bot;
  VALUE Top;
  BOOL flag;
  INT64 LatticeBot = TheLatticeBot;
  INT64 LatticeTop = TheLatticeTop;
  Bot.i = LatticeBot;
  Top.i = LatticeTop;
  REG reg;


  ClearBblValueSpecsFun(ifun);
 
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (ABBL_IS_KILLED(ibbl)) continue;
      if (BBL_CONSTANT_PROPAGATION)
	{
	  NewBblValueSpec(ibbl,Bot);
#if !PROP_DEAD_VALUES
	  REGSET_FOREACH_REG(~AbblRegsLiveIn(ibbl),reg)
	    SetOneBblValueSpec(ibbl,Top,reg);
	  SetOneBblValueSpec(ibbl,Bot,REG_GP);
	  SetOneBblValueSpec(ibbl,Bot,REG_ZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_FZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_RA);
	  SetOneBblValueSpec(ibbl,Bot,REG_SP);
	  SetOneBblValueSpec(ibbl,Bot,REG_PV);
	  SetOneBblValueSpec(ibbl,Bot,REG_AT);
#endif
	}
      else 
	{
	  NewBblValueSpec(ibbl,Top);
	  SetOneBblValueSpec(ibbl,Bot,REG_GP);
	  SetOneBblValueSpec(ibbl,Bot,REG_ZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_FZERO);
	  SetOneBblValueSpec(ibbl,Bot,REG_RA);
	  SetOneBblValueSpec(ibbl,Bot,REG_SP);
	  SetOneBblValueSpec(ibbl,Bot,REG_PV);
	  SetOneBblValueSpec(ibbl,Bot,REG_AT);
	}
      ABBL_UNMARK(ibbl);
      ABBL_UNLIVE(ibbl);
      flag = 0;
      FOREACH_BBL_PRED(ibbl,pred)
	{
	  if (AEDG_IS_INTER(pred) && AEDG_LIVE(pred))
	    {
	      
	      if (!ABBL_IS_FIRST(ibbl))
		{
		  JoinValueSpec(GetEdgValueSpec(pred),GetBblValueSpec(ibbl),
				AEDG_TYPE(pred)==ET_FUN_RETURN?
				(
				 /*AFUN_PSEUDO_PAL == ABBL_FUN(AEDG_HEAD(pred))?REG2BIT(REG_V0):*/AFUN_REGS_CHANGED(ABBL_FUN(AEDG_HEAD(pred))) // NIET 100% ZEKER
				 )
				:REGS_ALL
				,AbblRegsLiveIn(ibbl)
				);
		}
	      ABBL_SET_LIVE(ibbl);
	      ABBL_MARK(ibbl);
	    }
	  if (AEDG_TYPE(pred)==ET_FUN_LINK && !AEDG_IS_INTER(pred))
	    flag = 1;
	}
      if (flag)
	{
	  ABBL_UNMARK(ibbl);
	  ABBL_UNLIVE(ibbl);
	}
    }
  CopyValueSpec(GetEdgValueSpec(call),GetBblValueSpec(AFUN_BBL_FIRST(ifun)));
}

static INT64 JoinValues(VALUESPEC* spec, REG reg, INT64 old_value)
{
  INT64 new_value;
  INT64 LatticeBot = TheLatticeBot;
  INT64 LatticeTop = TheLatticeTop;  
  
  if (spec->bot & REG2BIT(reg))
    new_value = old_value;
  else if (spec->top & REG2BIT(reg))
    new_value = LatticeTop;
  else
    {
      if( old_value == LatticeTop )
	{
	  new_value = LatticeTop;
	}
      else if( old_value == LatticeBot )
	{
	  new_value = spec->value[reg].i;
	}
      else if( old_value == spec->value[reg].i)
	{
	  new_value = old_value;
	}
      else
	{
	  new_value = LatticeTop;
	}
    }
  
  return new_value;
  
}
/*=============================================================
  merge the temporary new and an old register sets for a bbl
  ===============================================================*/

static void AbblProvSetValueSpec(INDEX ibbl)
{
  INDEX iins;
  VALUESPEC* input = GetBblValueSpec(ibbl);
  VALUESPEC tmp1;
  VALUESPEC* output = &tmp1;//(VALUESPEC *) malloc(sizeof(VALUESPEC));
  REG reg;
  
  CopyValueSpec(input,output);
  
  FOREACH_BBL_INS(ibbl,iins)
    {
      if (AINS_FLAGS(iins) & F_KILLED) continue;
      
      reg = AINS_REG_A(iins);
      
      if (!(AINS_FLAGS(iins) & FI_EVAL_A))
	{
	  if ((AINS_TYPE(iins)!=IT_IST && AINS_TYPE(iins)!=IT_CST && AINS_TYPE(iins)!=IT_FST) || AINS_REG_B(iins)!=REG_SP)
	    AINS_A_EVAL(iins)=JoinValues(output,reg,AINS_A_EVAL(iins));
	  else
	    AINS_A_EVAL(iins)=TheLatticeTop;
	}

      reg = AINS_REG_B(iins);
      
      if (!(AINS_FLAGS(iins) & FI_EVAL_B))
	AINS_B_EVAL(iins)=JoinValues(output,reg,AINS_B_EVAL(iins));
      
      reg = AINS_REG_C(iins);
      
      AinsEvalValueSpecNew(iins,output);
      
      if (!(AINS_FLAGS(iins) & FI_EVAL_C))
	AINS_C_EVAL(iins)=JoinValues(output,reg,AINS_C_EVAL(iins));
    }
  //  free(output);
}


