/*=======================================================
  Author:   R. Muth
  Version:  $Id: alto.c,v 1.1.1.1 2000/06/07 12:37:02 brdsutte Exp $
  Comment:  This file contains the function main 
=======================================================*/

#include "global.h"
#include "functions.h"
#include "flags.h"
#include "regs.h"
#include "system.coff.h"

#include <unistd.h>
#include <sys/stat.h>

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "main"
char * exec_name=NULL;

void isignal(int i)
{
  on();
  DEBUG;
}

void jsignal(int i) // ALL KINDS OF STUFF THAT WE MIGHT DO DURING EVALUATION THAT THE PROGRAM UNDER COMPACTION CANNOT DO
{
}

#undef __PROC__
#define __PROC__ "main"

int main(int argc, char *argv[])
{
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constinfo");


    PrintSep0(stdout);
#ifndef SQUEEZE
    fprintf(stdout,
            "Alto Version 99/08/23\n"
            "For more information check http://www.cs.arizona.edu/alto/\n"
            "and the README file\n");
#else /* SQUEEZE */
    fprintf(stdout, "Squeeze Version 0.3.4-ghent (Jan 2001)\n");
    fprintf(stdout, "for more information check http://www.cs.arizona.edu/squeeze/\n");
#endif /* SQUEEZE */
    PrintSep0(stdout);

    exec_name=argv[0];

    ParseCommandLine(argc,argv);

    init_debug();
    signal(8,jsignal);

    EmulInitialize();
    
    BinaryLoad();

    BinaryDisassemble(); 

    BinaryConstructFlowGraph();

    BinaryTrivialOpts();

    if( HasOption(PRINT_OPT_WHEN,"Preproc") )
        PrintAll(stdout);

    BinaryProcess();

    if( HasOption(PRINT_OPT_WHEN,"Postproc") )
        PrintAll(stdout);

    if (verbose)
      {
	long cnt_zero=0;
	long cnt_code=0;
	long cnt_lita=0;
	long cnt_rconst=0;
	long cnt_data=0;
	long cnt_bss=0;
	long cnt_data_load=0;
	long cnt_bss_load=0;
	long cnt_data_other=0;
	long cnt_bss_other=0;
	long cnt_other=0;
	long cnt_small=0;
	long cnt_all=0;
	INDEX iins;
	long nr_iins =0L;
	FOREACH_LIVE_INS(iins)
	  {
	    nr_iins++;
	    if (AINS_FLAGS(iins) & FI_EVAL_C)
	      {
		if (AINS_REG_C(iins)==REG_ZERO ) continue;
	    
		cnt_all++;

		if (AINS_C_EVAL(iins)==0)
		  {
		    cnt_zero++;
		  }

		else if (AddressIsCode(AINS_C_EVAL(iins)))
		  cnt_code++;
		else if (AsecContainsAddress(BIN_SECTION_MAP[R_SN_LITA],AINS_C_EVAL(iins)))
		  cnt_lita++;
		else if (AsecContainsAddress(BIN_SECTION_MAP[R_SN_RCONST],AINS_C_EVAL(iins)))
		  cnt_rconst++;
		else if ((AINS_C_EVAL(iins)>=BIN_AOUTHDR.data_start) && (AINS_C_EVAL(iins)<=(BIN_AOUTHDR.data_start+BIN_AOUTHDR.dsize)))
		  {
		    cnt_data++; 
		    if ((AINS_TYPE(iins)==IT_ILD)||(AINS_TYPE(iins)==IT_ULD)) cnt_data_load++;
		    else  cnt_data_other++;
		  }
		else if ((AINS_C_EVAL(iins)>=BIN_AOUTHDR.bss_start) && (AINS_C_EVAL(iins)<=(BIN_AOUTHDR.data_start+BIN_AOUTHDR.bsize)))
		  {
		    cnt_bss++; 
		    if ((AINS_TYPE(iins)==IT_ILD)||(AINS_TYPE(iins)==IT_ULD)) cnt_bss_load++;
		    else cnt_bss_other++;
		  }

		else if (AINS_C_EVAL(iins) <255 && AINS_C_EVAL(iins)>-255)
		  cnt_small++; 
		else
		  cnt_other++;
	      }
	  }

	printf("Found %5ld zero constants\n",cnt_zero);
	printf("Found %5ld small constants\n",cnt_small);
	printf("Found %5ld code constants\n",cnt_code);
	printf("Found %5ld lita constants\n",cnt_lita);
	printf("Found %5ld rconst constants\n",cnt_rconst);
	printf("Found %5ld data constants    (%ld load / %ld other)\n",cnt_data,cnt_data_load,cnt_data_other);
	printf("Found %5ld bss constants     (%ld load / %ld other)\n",cnt_bss,cnt_bss_load,cnt_bss_other);
	printf("Found %5ld other constants\n",cnt_other);
	printf("----- %5ld constants, %ld ins\n",cnt_all,nr_iins);
      }

#if 0
    OptRemoveUnreachableCode(stdout);
    ComputeSavedChangedRegisters(stdout);

    ComUseDef(stdout,1);
    ComInsDefPotential(stdout,1);
#endif
    
    if( HasOption(PRINT_OPT_WHEN,"Preasm") )
        PrintAll(stdout);

    BinaryAssemble();
    
    if( HasOption(PRINT_OPT_WHEN,"Postasm") )
        PrintAll(stdout);
        
    if( HasOption(PRINT_OPT_WHEN,"Presave") )
        PrintAll(stdout);

    unlink(CONFIG_OUTFILE);
    BinarySave();
    chmod(CONFIG_OUTFILE, S_IXUSR | S_IRUSR | S_IWUSR);
    
    if( HasOption(PRINT_OPT_WHEN,"Postsave") )
        PrintAll(stdout);

    return 0;
}

/*=======================================================
=======================================================*/




