#include "global.h"
#include "flags.h"
#include "regs.h"
#include "opcode.h"
#include "functions.h"
#include "blocks.h"

VALUESPEC valuespec[CONFIG_NSPEC];
int nr_edg_specs;
int nr_bbl_specs;
int nr_intra_edg_specs;

extern INT64 TheLatticeTop;
extern INT64 TheLatticeBot;

/**********************************************************
 general valuespec functions
 **********************************************************/

extern void ClearEdgValueSpecs()
{
  INDEX edge;

  FOREACH_EDG(edge) 
    AEDG_VALUESPEC(edge) = -1;
  nr_edg_specs = 0;
}

extern void ClearBblValueSpecs()
{
  INDEX ibbl;

  FOREACH_BBL(ibbl) 
    ABBL_VALUESPEC(ibbl) = -1;
  nr_bbl_specs = SPEC_FIRST_BBL;
}

extern void ClearIntraEdgValueSpecs()
{
  INDEX ibbl;
  INDEX iedg;

  FOREACH_LIVE_EDG(iedg)
    if (!AEDG_IS_INTER(iedg))
      AEDG_VALUESPEC(iedg) = -1;
  nr_intra_edg_specs = SPEC_FIRST_INTRA_EDG;
}

extern void ClearBblValueSpecsFun(INDEX ifun)
{
  INDEX ibbl;

  FOREACH_FUN_BBL(ifun,ibbl) 
    ABBL_VALUESPEC(ibbl) = -1;
  nr_bbl_specs = SPEC_FIRST_BBL;
}


extern void ClearIntraEdgValueSpecsFun(INDEX ifun)
{
  INDEX ibbl;
  INDEX iedg;

  FOREACH_FUN_BBL(ifun,ibbl) 
    FOREACH_BBL_SUCC(ibbl,iedg)
    if (!AEDG_IS_INTER(iedg))
      AEDG_VALUESPEC(iedg) = -1;
  nr_intra_edg_specs = SPEC_FIRST_INTRA_EDG;
}

extern void SetOneBblValueSpec(INDEX ibbl, VALUE value, REG reg)
{
  if (!VALID(ABBL_VALUESPEC(ibbl))) return;
  valuespec[ABBL_VALUESPEC(ibbl)].value[reg].i = value.i;
  if (value.i == TheLatticeBot)
    {
      valuespec[ABBL_VALUESPEC(ibbl)].bot |= REG2BIT(reg);
      valuespec[ABBL_VALUESPEC(ibbl)].top &= ~REG2BIT(reg);
    }
  else if (value.i == TheLatticeTop)
    {
      valuespec[ABBL_VALUESPEC(ibbl)].bot &= ~REG2BIT(reg);
      valuespec[ABBL_VALUESPEC(ibbl)].top |= REG2BIT(reg);
    }
  else 
    {
      valuespec[ABBL_VALUESPEC(ibbl)].bot &= ~REG2BIT(reg);
      valuespec[ABBL_VALUESPEC(ibbl)].top &= ~REG2BIT(reg);
    }
}     

extern void SetOneEdgValueSpec(INDEX iedg, VALUE value, REG reg)
{
  if (!VALID(AEDG_VALUESPEC(iedg))) return;
  valuespec[AEDG_VALUESPEC(iedg)].value[reg].i = value.i;
  if (value.i == TheLatticeBot)
    {
      valuespec[AEDG_VALUESPEC(iedg)].bot |= REG2BIT(reg);
      valuespec[AEDG_VALUESPEC(iedg)].top &= ~REG2BIT(reg);
    }
  else if (value.i == TheLatticeTop)
    {
      valuespec[AEDG_VALUESPEC(iedg)].bot &= ~REG2BIT(reg);
      valuespec[AEDG_VALUESPEC(iedg)].top |= REG2BIT(reg);
    }
  else 
    {
      valuespec[AEDG_VALUESPEC(iedg)].bot &= ~REG2BIT(reg);
      valuespec[AEDG_VALUESPEC(iedg)].top &= ~REG2BIT(reg);
    }
}     

extern BOOL NewEdgValueSpec(INDEX iedg,VALUE value)
{
  REG reg;
  VALUE Zero,Fzero;
  Zero.i = 0L;
  Fzero.f = 0.0;
  if ( nr_edg_specs == SPEC_FIRST_BBL) 
    {
      fprintf(stdout,"Not enough Edge ValueSpecs available\n");
      return 0;
    }
  AEDG_VALUESPEC(iedg)=nr_edg_specs;
  if (value.i==TheLatticeBot)
    {
      valuespec[AEDG_VALUESPEC(iedg)].bot = REGS_ALL;
      valuespec[AEDG_VALUESPEC(iedg)].top = REGS_NONE;
    }
  else if (value.i==TheLatticeTop)
    {
      valuespec[AEDG_VALUESPEC(iedg)].top = REGS_ALL;
      valuespec[AEDG_VALUESPEC(iedg)].bot = REGS_NONE;
    }
  else for (reg=0;reg<64;reg++)
    valuespec[AEDG_VALUESPEC(iedg)].value[reg].i = value.i;
  SetOneEdgValueSpec(iedg,Zero,REG_ZERO);
  SetOneEdgValueSpec(iedg,Fzero,REG_FZERO);
  nr_edg_specs++;
  return 1;
} 

/********************************************************
 reuse a valuespec of an edge to avoid overflow on the
 valuespec array
 ********************************************************/

extern BOOL ReuseEdgValueSpec(INDEX iedg,INT32 spec, VALUE value)
{
  REG reg;
  VALUE Zero,Fzero;
  Zero.i = 0L;
  Fzero.f = 0.0;
  AEDG_VALUESPEC(iedg)=spec;
  if (value.i==TheLatticeBot)
    {
      valuespec[AEDG_VALUESPEC(iedg)].bot = REGS_ALL;
      valuespec[AEDG_VALUESPEC(iedg)].top = REGS_NONE;
    }
  else if (value.i==TheLatticeTop)
    {
      valuespec[AEDG_VALUESPEC(iedg)].top = REGS_ALL;
      valuespec[AEDG_VALUESPEC(iedg)].bot = REGS_NONE;
    }
  else for (reg=0;reg<64;reg++)
    valuespec[AEDG_VALUESPEC(iedg)].value[reg].i = value.i;
  SetOneEdgValueSpec(iedg,Zero,REG_ZERO);
  SetOneEdgValueSpec(iedg,Fzero,REG_FZERO);
  return 1;
} 

extern BOOL NewBblValueSpec(INDEX ibbl,VALUE value)
{
  REG reg;
  VALUE Zero,Fzero;
  Zero.i = 0L;
  Fzero.f = 0.0;
  if ( nr_bbl_specs == SPEC_FIRST_INTRA_EDG)
    {
      fprintf(stdout,"Not enough Bbl ValueSpecs available\n");
      return 0;
    }
  ABBL_VALUESPEC(ibbl) = nr_bbl_specs;
  if (value.i==TheLatticeBot)
    {
      valuespec[ABBL_VALUESPEC(ibbl)].bot = REGS_ALL;
      valuespec[ABBL_VALUESPEC(ibbl)].top = REGS_NONE;
    }
  else if (value.i==TheLatticeTop)
    {
      valuespec[ABBL_VALUESPEC(ibbl)].top = REGS_ALL;
      valuespec[ABBL_VALUESPEC(ibbl)].bot = REGS_NONE;
    }
  else for (reg=0;reg<64;reg++)
    valuespec[ABBL_VALUESPEC(ibbl)].value[reg].i = value.i;
  
  SetOneBblValueSpec(ibbl,Zero,REG_ZERO);
  SetOneBblValueSpec(ibbl,Fzero,REG_FZERO);
  nr_bbl_specs++;
  return 1;
}

extern BOOL NewIntraEdgValueSpec(INDEX iedg,VALUE value)
{
  REG reg;
  VALUE Zero,Fzero;
  Zero.i = 0L;
  Fzero.f = 0.0;
  if ( nr_intra_edg_specs == CONFIG_NSPEC)
    {
      fprintf(stdout,"Not enough Bbl ValueSpecs available\n");
      return 0;
    }
  AEDG_VALUESPEC(iedg) = nr_intra_edg_specs;
  if (value.i==TheLatticeBot)
    {
      valuespec[AEDG_VALUESPEC(iedg)].bot = REGS_ALL;
      valuespec[AEDG_VALUESPEC(iedg)].top = REGS_NONE;
    }
  else if (value.i==TheLatticeTop)
    {
      valuespec[AEDG_VALUESPEC(iedg)].top = REGS_ALL;
      valuespec[AEDG_VALUESPEC(iedg)].bot = REGS_NONE;
    }
  else for (reg=0;reg<64;reg++)
    valuespec[AEDG_VALUESPEC(iedg)].value[reg].i = value.i;
  
  SetOneEdgValueSpec(iedg,Zero,REG_ZERO);
  SetOneEdgValueSpec(iedg,Fzero,REG_FZERO);
  nr_intra_edg_specs++;
  return 1;
}

extern INT64 GetOneBblValueSpec(INDEX ibbl, REG reg)
{
  if (!VALID(ABBL_VALUESPEC(ibbl)))
    {
      fprintf(stderr,"value asked form non-existent bbl valuespec\n");
      return 0;
    }  
  return valuespec[ABBL_VALUESPEC(ibbl)].value[reg].i;
}

extern INT64 GetOneEdgValueSpec(INDEX iedg, REG reg)
{
  if (!VALID(AEDG_VALUESPEC(iedg)))
    {
      fprintf(stderr,"value asked form non-existent edg valuespec\n");
      return 0;
    }  
  return valuespec[AEDG_VALUESPEC(iedg)].value[reg].i;
}

extern VALUESPEC* GetBblValueSpec(INDEX ibbl)
{
  if (!VALID(ABBL_VALUESPEC(ibbl)))
    return NULL;
  return &(valuespec[ABBL_VALUESPEC(ibbl)]);
}

extern VALUESPEC* GetEdgValueSpec(INDEX iedg)
{
  if (!VALID(AEDG_VALUESPEC(iedg)))
    return NULL;
  return &valuespec[AEDG_VALUESPEC(iedg)];
}

extern BOOL JoinAllValueSpec(VALUESPEC* source,VALUESPEC* dest,REGSET regs_live)
{
  return JoinValueSpec(source,dest,REGS_ALL,regs_live);
}

extern int super_change;
#define AEDG_LIVE(ibbl)      (AEDG_FLAGS(ibbl) &  FE_LIVE)
#define AEDG_UNLIVE(ibbl)     AEDG_FLAGS(ibbl) &= ~FE_LIVE
#define AEDG_SET_LIVE(ibbl)   AEDG_FLAGS(ibbl) |= FE_LIVE

extern BOOL JoinValueSpec(VALUESPEC* source,VALUESPEC* dest, REGSET regs, REGSET regs_live)
{
  REG reg;
  INT64 old_value,value;
  BOOL change;
  MAP_BLOCK * block;
  INT64 dest_top = dest->top;
  INT64 dest_bot = dest->bot;
  INT64 source_top  = source->top;
  INT64 source_bot  = source->bot;
  INT64 bitreg;
  INT64 old_top = dest_top;
  REGSET prelimint;
  REGSET prelimflt;

  /* 
     merges source and dest in dest and 
     returnes 0 if nothing changed to dest, 
     returnes 1 otherwise
  */
  
  change = 0;

  prelimint = REGS_INT & regs_live & regs & ~source_bot & ~dest_bot & ((dest_top & ~source_top) | (~dest_top & source_top));

  if (prelimint)
    {
      REGSET_FOREACH_INT_REG(prelimint, reg)
	{
	  ADDRESS address;
	  if (REG2BIT(reg) & (dest_top & ~source_top))
	    address = source->value[reg].i;
	  else
	    address = dest->value[reg].i;
	  if (!AddressIsCode(address))
	    {	  
	      //	  block=LocateBlock(address);
	      //	  if (block)
	      AdaptMapOTF(BLOCK_STORE,address,0);
	    }
	  else
	    {
	      // set edge pointed to by address at location live
	      INDEX ifun=ABBL_FUN(AINS_BBL(Addr2Ins(address)));
	      INDEX bbl=AFUN_BBL_FIRST(ifun);
	      INDEX iedg;
	      AFUN_FLAGS(ifun) |= F_MARKED;
	      FOREACH_BBL_PRED(bbl,iedg)
		if (ABBL_FUN(AEDG_HEAD(iedg))==AFUN_PSEUDO_HELL_LITE)
		  {
		    AEDG_MARK(iedg);
		    if (!AEDG_LIVE(iedg))
		      {
			super_change=1;
			//			printf("2 making edge to %d %s live\n",ifun,AFUN_NAME(ifun));
		      }
		    AEDG_SET_LIVE(iedg);
		  }
	    }
	}
    }

  prelimflt = REGS_FLT & regs_live & regs & ~source_bot & ~dest_bot & ((dest_top & ~source_top) | (~dest_top & source_top));

  if (prelimflt)
    {
      REGSET_FOREACH_FLT_REG(prelimflt, reg)
	{
	  ADDRESS address;
	  if (REG2BIT(reg) & (dest_top & ~source_top))
	    address = source->value[reg].i;
	  else
	    address = dest->value[reg].i;
	  if (!AddressIsCode(address))
	    {	  
	      //	  block=LocateBlock(address);
	      //	  if (block)
	      AdaptMapOTF(BLOCK_STORE,address,0);
	    }
	  else
	    {
	      // set edge pointed to by address at location live
	      INDEX ifun=ABBL_FUN(AINS_BBL(Addr2Ins(address)));
	      INDEX bbl=AFUN_BBL_FIRST(ifun);
	      INDEX iedg;
	      AFUN_FLAGS(ifun) |= F_MARKED;
	      FOREACH_BBL_PRED(bbl,iedg)
		if (ABBL_FUN(AEDG_HEAD(iedg))==AFUN_PSEUDO_HELL_LITE)
		  {
		    AEDG_MARK(iedg);
		    if (!AEDG_LIVE(iedg))
		      {
			super_change=1;
			printf("3 making edge to %d %s live\n",ifun,AFUN_NAME(ifun));
		      }
		    AEDG_SET_LIVE(iedg);
		  }
	    }
	}
    }

  dest_top |= (source_top & regs);
  dest_bot &= ~(source_top & regs);
  
  change = (dest_top != old_top);
  
  regs &= ~(dest_top);
  regs &= ~(source_bot);
  regs &= REGS_ALL_BUT_ZERO;
  
  if (regs & REGS_INT_BUT_ZERO)
    {
      REGSET_FOREACH_INT_BITREG(regs,reg,bitreg)
	{
	  value = source->value[reg].i;
	  
	  if (dest_bot & bitreg)
	    {
	      dest->value[reg].i = value;
	      dest_bot &= ~bitreg;
	      dest_top &= ~bitreg;
	      change = 1;
	      continue;
	    }
	  
	  old_value = dest->value[reg].i;
	  if ( old_value == value )
	    continue;
	  
	  dest_top |= bitreg;
	  dest_bot &= ~bitreg;
	  
	  if (bitreg & regs_live)
	    {
	      ADDRESS address = source->value[reg].i;
	      if (!AddressIsCode(address))
		AdaptMapOTF(BLOCK_STORE,address,0);

	      address = old_value;
	      if (!AddressIsCode(address))
		AdaptMapOTF(BLOCK_STORE,address,0);
	    }
	  
	  change = 1;
	}
    }

  if (regs & REGS_FLT_BUT_ZERO)
    {
      REGSET_FOREACH_FLT_BITREG(regs,reg,bitreg)
	{
	  value = source->value[reg].i;
	  
	  if (dest_bot & bitreg)
	    {
	      dest->value[reg].i = value;
	      dest_bot &= ~bitreg;
	      dest_top &= ~bitreg;
	      change = 1;
	      continue;
	    }
	  
	  old_value = dest->value[reg].i;
	  if ( old_value == value )
	    continue;
	  
	  dest_top |= bitreg;
	  dest_bot &= ~bitreg;

	  if (bitreg & regs_live)
	    {
	      ADDRESS address = source->value[reg].i;
	      if (!AddressIsCode(address))
		AdaptMapOTF(BLOCK_STORE,address,0);
	      
	      address = old_value;
	      if (!AddressIsCode(address))
		AdaptMapOTF(BLOCK_STORE,address,0);
	    }	
  
	  change = 1;
	}
    }

  dest->top = dest_top; 
  dest->bot = dest_bot;
  return change;
}

extern void CopyValueSpec(VALUESPEC* source, VALUESPEC* dest)
{
  memcpy(dest,source,sizeof(VALUESPEC));
}

extern void PrintValueSpec(VALUESPEC* spec,REGSET regs)
{
  int i;
  printf("valuespec:\n");
  REGSET_FOREACH_REG(regs & ~spec->top & ~spec->bot,i)
    printf("reg %d: %lx\n",i,spec->value[i]);
  printf("Bot: ");
  PrintRegSet(regs & spec->bot);
  printf("\n");
  printf("Top: ");
  PrintRegSet(regs & spec->top);
  printf("\n");
}
